"""Utility functions and classes to manage information on EU countries."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import PIL
from PIL import Image, ImageDraw, ImageFont
import requests
from io import BytesIO
from urllib.request import urlopen
import json


#########################################################################################################################################
# EU countries code taken from:
# https://ec.europa.eu/eurostat/statistics-explained/index.php?title=Glossary:Country_codes
#########################################################################################################################################


# Conversion from iso_a2 country codes to iso_a3 country codes (needed for creating plotly choropleth map)
with urlopen('https://raw.githubusercontent.com/flyingcircusio/pycountry/master/src/pycountry/databases/iso3166-1.json') as response:
    pycountry = json.load(response)

iso2iso3 = {}
iso3iso2 = {}
iso2name = {}
for c in pycountry['3166-1']:
    iso2iso3[c['alpha_2']] = c['alpha_3']
    iso3iso2[c['alpha_3']] = c['alpha_2']
    iso2name[c['alpha_2']] = c['name']

iso2iso3['UK'] = 'GBR'
iso2name['UK'] = 'United Kingdom'

iso2iso3['EU'] = 'EU'
iso2name['EU'] = 'European Union'

iso2iso3['EA'] = 'EA'
iso2name['EA'] = 'Euro Area'

iso2iso3['EU28'] = 'EU28'
iso2name['EU28'] = 'European Union'

iso2iso3['EL'] = 'GRC'
iso2name['EL'] = 'Greece'


# Base URL of the EU flags
baseurlflags = 'https://jeodpp.jrc.ec.europa.eu/services/shared/Notebooks/images/flags/'
    
    
# country class
class country:
    """
    Class to store information on a single EU country.
    
    Parameters
    ----------
        name : str
            Name of the country
        iso2code : str
            Two letter code of the country defined by EUROSTAT (https://ec.europa.eu/eurostat/statistics-explained/index.php?title=Glossary:Country_codes)
        iso3code : str
            Three letter code of the country defined by ISO 3166 (https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3)
        euro : bool, optional
            Flag that is True for countries belonging to the Euro Area (default is False)
        iscountry : bool, optional
            True if it is a real country (default is True). For instance 'Euro Area' and 'European Union' have this value set to False
        population: int, optional
            Last known population of the country
    """

    def __init__(self, name, iso2code, euro=False, iscountry=True, population=0):
        self.name       = name
        self.iso2code   = iso2code
        self.iso3code   = iso2iso3[iso2code]
        self.euro       = euro
        self.iscountry  = iscountry
        self.population = population
        self.flag       = baseurlflags + iso2code + '.png'
        
        
    def __str__(self):
        """Returns the country name as __str__"""
        return self.name
    
    def __repr__(self):
        """Returns the country name as __repr__"""
        return self.name
    
    def flagImage(self):
        """Returns the flag of the country as a PIL.Image object"""
        response = requests.get(self.flag)
        img = Image.open(BytesIO(response.content))
        return img        
    
    
    
# eucountries class            
class countries:
    """
    Class to store information on all the EU countries.
    
    Parameters
    ----------
        countries_list : list of eucountries.country instances
            List of countries in the European Union
            
    Examples
    --------
    Get the list of all European Union or Euro Area countries::
    
        from vois import eucountries as eu

        countriesEU = eu.countries.EuropeanUnion()
        print(countriesEU)
        
        countriesEuro = eu.countries.EuroArea()
        print(countriesEuro)
        
    
    Display the flag image of a country given its code::
    
        from vois import eucountries as eu
        display( eu.countries.byCode('IT').flagImage() )
        
        
    Display the flag image of a country given its name::
    
        from vois import eucountries as eu
        display( eu.countries.byName('Lithuania').flagImage() )
        
    .. figure:: figures/flag.png
       :scale: 100 %
       :alt: Flag of a country

       Display of the flag of an EU country
    
    Get the list of all the European Union country codes::
    
        from vois import eucountries as eu
        display( eu.countries.EuroAreaCodes() )
        
    """
    
    countries_list = []
    
    @classmethod
    def add(cls, name, iso2code, euro=False, iscountry=True, population=0):
        """
        Static method to add a country to the countries_list
    
        Parameters
        ----------
        name : str
            Name of the country
        iso2code : str
            Two letter code of the country defined by EUROSTAT (https://ec.europa.eu/eurostat/statistics-explained/index.php?title=Glossary:Country_codes)
        euro : bool, optional
            Flag that is True for countries belonging to the Euro Area (default is False)
        iscountry : bool, optional
            True if it is a real country (default is True). For instance 'Euro Area' and 'European Union' have this value set to False
        population: int, optional
            Last known population of the country
        """
        cls.countries_list.append( country(name, iso2code, euro=euro, iscountry=iscountry, population=population))
    
    
    
    # Returns the list of all the country belonging to the Euro Area
    @classmethod
    def EuroArea(cls, sortByName=True):
        """
        Static method that returns the list of all the country belonging to the Euro Area
        
        Parameters
        ----------
            sortByName : bool, optional
                If True, the returned list of countries is sorted by the name of the country (default is True)
        """
        res = []
        for c in cls.countries_list:
            if c.iscountry and c.euro:
                res.append(c)
        if sortByName:
            return sorted(res, key=lambda x: x.name, reverse=False)
        else:
            return sorted(res, key=lambda x: x.iso2code, reverse=False)

        
    # Returns the list of all the iso2codes of the countries belonging to the Euro Area
    @classmethod
    def EuroAreaCodes(cls):
        """
        Static method that returns the list of all the iso2codes of the countries belonging to the Euro Area
        """
        countries = cls.EuroArea()
        return [c.iso2code for c in countries]


    
    # Returns the list of all the country belonging to the European Union
    @classmethod
    def EuropeanUnion(cls, sortByName=True):
        """
        Static method that returns the list of all the country belonging to the European Union
        
        Parameters
        ----------
            sortByName : bool, optional
                If True, the returned list of countries is sorted by the name of the country (default is True)
        """
        res = []
        for c in cls.countries_list:
            if c.iscountry:
                res.append(c)
        if sortByName:
            return sorted(res, key=lambda x: x.name, reverse=False)
        else:
            return sorted(res, key=lambda x: x.iso2code, reverse=False)

        
    # Returns the list of all the iso2codes of the countries belonging to the European Union
    @classmethod
    def EuropeanUnionCodes(cls):
        """
        Static method that returns the list of all the iso2codes of the countries belonging to the European Union
        """
        countries = cls.EuropeanUnion()
        return [c.iso2code for c in countries]
        

    # Returns the list of all the names of the countries belonging to the European Union
    @classmethod
    def EuropeanUnionNames(cls):
        """
        Static method that returns the list of all the names of the countries belonging to the European Union
        """
        countries = cls.EuropeanUnion()
        return [c.name for c in countries]
    
    
    @classmethod
    def byCode(cls, iso2code):
        """
        Static method that returns a country given its iso2code, or None if not existing
        
        Parameters
        ----------
            iso2code : str
                Two letter code of the country defined by EUROSTAT (https://ec.europa.eu/eurostat/statistics-explained/index.php?title=Glossary:Country_codes)
                
        Returns
        -------
            Instance of country or None
            
        """
        res = [c for c in cls.countries_list if c.iso2code == iso2code]
        if len(res) > 0: return res[0]
                
    @classmethod
    def byName(cls, name):
        """
        Static method that returns a country given its name, or None if not existing
        
        Parameters
        ----------
            name : str
                Name of the country
                
        Returns
        -------
            Instance of country or None
            
        """
        res = [c for c in cls.countries_list if c.name == name]
        if len(res) > 0: return res[0]
            

            
# Population data for 2018 obtained from https://data.worldbank.org/indicator/SP.POP.TOTL CSV file
eu_total_population = 446786293

countries.add('Belgium',     'BE',euro=True,population=11433256)
countries.add('Greece',      'EL',euro=True,population=10731726)
countries.add('Lithuania',   'LT',euro=True,population=2801543)
countries.add('Portugal',    'PT',euro=True,population=10283822)
countries.add('Bulgaria',    'BG',population=7025037)
countries.add('Spain',       'ES',euro=True,population=46796540)
countries.add('Luxembourg',  'LU',euro=True,population=607950)
countries.add('Romania',     'RO',population=19466145)
countries.add('Czechia',     'CZ',population=10629928)
countries.add('France',      'FR',euro=True,population=66977107)
countries.add('Hungary',     'HU',population=9775564)
countries.add('Slovenia',    'SI',euro=True,population=2073894)
countries.add('Denmark',     'DK',population=5793636)
countries.add('Croatia',     'HR',euro=True,population=4087843)
countries.add('Malta',       'MT',euro=True,population=484630)
countries.add('Slovakia',    'SK',euro=True,population=5446771)
countries.add('Germany',     'DE',euro=True,population=82905782)
countries.add('Italy',       'IT',euro=True,population=60421760)
countries.add('Netherlands', 'NL',euro=True,population=17231624)
countries.add('Finland',     'FI',euro=True,population=5515525)
countries.add('Estonia',     'EE',euro=True,population=1321977)
countries.add('Cyprus',      'CY',euro=True,population=1189265)
countries.add('Austria',     'AT',euro=True,population=8840521)
countries.add('Sweden',      'SE',population=10175214)
countries.add('Ireland',     'IE',euro=True,population=4867309)
countries.add('Latvia',      'LV',euro=True,population=1927174)
countries.add('Poland',      'PL',population=37974750)

#countries.add('United Kingdom', 'UK')

countries.add('European Union','EU', iscountry=False, population=eu_total_population)
countries.add('Euro Area',     'EA', iscountry=False, population=341858176+4087843)


# language class
class language:
    """
    Class to store information on a language of the European Union.
    
    Parameters
    ----------
        name : str
            Name of the language
        abbreviation: str
            Abbreviation used for the language
        population : int, optional
            Last known population that uses the language
                
    """

    def __init__(self, name, abbreviation, population=0):
        self.name         = name
        self.abbreviation = abbreviation
        self.population   = population
        
    def __str__(self):
        """Returns the language name as __str__"""
        return self.name
    
    def __repr__(self):
        """Returns the language name as __repr__"""
        return self.name

    
# language class            
class languages:
    """
    Class to store information on all the European Union languages.
    
    Parameters
    ----------
        languages_list : list of eucountries.language instances
            List of languages in the European Union
            
    Examples
    --------
    Print the list of all European Union languages::
    
        from vois import eucountries as eu
        names = eu.languages.EuropeanUnionLanguages(sortByName=False)
        print(names)
        
    Print the list of all European Union languages abbreviations::
    
        from vois import eucountries as eu
        abbrev = eu.languages.EuropeanUnionAbbreviations()
        print(abbrev)
        
    """
    
    languages_list = []
    
    @classmethod
    def add(cls, name, abbreviation, population=0):
        cls.languages_list.append( language(name, abbreviation, population) )
    

    
    # Returns the list of all the official languages to the European Union
    @classmethod
    def EuropeanUnionLanguages(cls, sortByName=True):
        """
        Static method that returns the list of all the languages of the European Union
        
        Parameters
        ----------
            sortByName : bool, optional
                If True, the returned list of languages is sorted by the name of the language (default is True)
        """
        if sortByName:
            return sorted(cls.languages_list, key=lambda x: x.name, reverse=False)
        else:
            return sorted(cls.languages_list, key=lambda x: x.abbreviation, reverse=False)

        
    # Returns the list of all the abbreviations of the official languages to the European Union
    @classmethod
    def EuropeanUnionAbbreviations(cls):
        """
        Static method that returns the list of all the abbreviations of the EU languages
        """
        return sorted([lang.abbreviation for lang in cls.languages_list])
        
        
    @classmethod
    def byAbbreviation(cls, abbreviation):
        """
        Static method that returns a language given its abbreviation, or None if not existing
        
        Parameters
        ----------
            abbreviation : str
                Two letter code of the language
                
        Returns
        -------
            Instance of language class or None
            
        """
        res = [c for c in cls.languages_list if c.abbreviation == abbreviation]
        if len(res) > 0: return res[0]
                
    @classmethod
    def byName(cls, name):
        """
        Static method that returns a language given its name, or None if not existing
        
        Parameters
        ----------
            name : str
                Name of the language
                
        Returns
        -------
            Instance of language class or None
            
        """
        res = [c for c in cls.languages_list if c.name == name]
        if len(res) > 0: return res[0]


# Languages population from https://en.wikipedia.org/wiki/Languages_of_the_European_Union
languages.add('Bulgarian',    'bg', int(0.02*eu_total_population) )
languages.add('Czech',        'cs', int(0.03*eu_total_population) )
languages.add('German',       'de', int(0.36*eu_total_population) )
languages.add('Greek',        'el', int(0.03*eu_total_population) )
languages.add('English',      'en', int(0.44*eu_total_population) )
languages.add('Spanish',      'es', int(0.17*eu_total_population) )
languages.add('French',       'fr', int(0.29*eu_total_population) )
languages.add('Croatian',     'hr', int(0.01*eu_total_population) )
languages.add('Hungarian',    'hu', int(0.03*eu_total_population) )
languages.add('Italian',      'it', int(0.18*eu_total_population) )
languages.add('Lithuanian',   'lt', int(0.01*eu_total_population) )
languages.add('Dutch',        'nl', int(0.06*eu_total_population) )
languages.add('Polish',       'pl', int(0.10*eu_total_population) )
languages.add('Portuguese',   'pt', int(0.03*eu_total_population) )
languages.add('Romanian',     'ro', int(0.06*eu_total_population) )
languages.add('Slovene',      'sl', int(0.005*eu_total_population) )
languages.add('Swedish',      'sv', int(0.03*eu_total_population) )

# Languages NOT found: check abbreviation!
languages.add('Danish',       'dk', int(0.01*eu_total_population) )
languages.add('Estonian',     'ee', int(0.005*eu_total_population) )
languages.add('Finnish',      'fi', int(0.01*eu_total_population) )
languages.add('Irish',        'ie', int(0.005*eu_total_population) )
languages.add('Latvian',      'lv', int(0.005*eu_total_population) )
languages.add('Maltese',      'mt', int(0.005*eu_total_population) )
languages.add('Slovak',       'sk', int(0.02*eu_total_population) )
        


"""Utilities functions for maps"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

import ipyleaflet
from ipyleaflet import basemaps, basemap_to_tiles, WidgetControl
from PIL import Image, ImageDraw
from io import BytesIO
import ipyvuetify
import ipyvuetify as v

from osgeo import ogr, osr
from jeodpp import inter       # V.I.: importing inter after GDAL seem to solve the HDF5 C errors

import requests
import math


# Symbols dimension in pixels
SMALL_SYMBOLS_DIMENSION = 30
MEDIUM_SYMBOLS_DIMENSION = 80
LARGE_SYMBOLS_DIMENSION = 256

# Base URL for generating images from symbols
SYMBOLS_URL = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view?x=0&y=0&z=1&procid=%s'


#####################################################################################################################################################
# Management of basemaps and layers (see viewer.py of Lucas)
#####################################################################################################################################################

# Build a basemap that works in old and new versions of ipyleaflet
def buildBasemap(oldstyledictionary):
    return basemap_to_tiles(oldstyledictionary)


def EmptyBasemap():  # gray background
    return buildBasemap({'attribution': '',
                         'max_zoom': 21, 
                         'max_native_zoom': 21,
                         'name': 'Empty basemap',
                         'url': 'https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/gray.png'})


def OSM_EC():
    return buildBasemap({'attribution': '',
                         'max_zoom': 21,
                         'max_native_zoom': 21,
                         'name': 'Gisco.OSMCartoV4Composite',
                         'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoV4Composite/EPSG3857/{z}/{x}/{y}.png'})


def CartoLabels():
    return buildBasemap({'attribution': '',
                         'max_zoom': 21,
                         'max_native_zoom': 21,
                         'name': 'Gisco.OSMCartoV4Labels',
                         'url': 'https://gisco-services.ec.europa.eu/maps/tiles/OSMCartoV4Labels/EPSG3857/{z}/{x}/{y}.png'})


def EsriWorldImagery():
    return buildBasemap({'attribution': '',
                         'max_zoom': 21,
                         'max_native_zoom': 21,
                         'name': 'Esri.WorldImagery',
                         'url': 'https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}.jpg'})


def GoogleRoadmap():
    return buildBasemap({'attribution': '',
                         'max_zoom': 21,
                         'max_native_zoom': 21,
                         'name': 'Google.Roadmap',
                         'url': 'https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}'})


def GoogleSatellite():
    return buildBasemap({'attribution': '',
                         'max_zoom': 21,
                         'max_native_zoom': 21,
                         'name': 'Google.Satellite',
                         'url': 'https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}'})


def GoogleHybrid():
    return buildBasemap({'attribution': '',
                         'max_zoom': 21,
                         'max_native_zoom': 21,
                         'name': 'Google.Hybrid',
                         'url': 'https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}'})


# Any ImageProcess or VectorLayer: returns an ipyleaflet.TileLayer instance
def BDAPLayer(p):
    procid = p.toLayer()
    #return ipyleaflet.TileLayer(url='https://jeodpp.jrc.ec.europa.eu/jiplib-view-dev?x={x}&y={y}&z={z}&procid=%s'%procid, max_zoom=21, max_native_zoom=21)
    return ipyleaflet.TileLayer(url='https://jeodpp.jrc.ec.europa.eu/jiplib-view?x={x}&y={y}&z={z}&procid=%s'%procid, max_zoom=21, max_native_zoom=21)


# Add the layer to a map or substitute it using the name
def addLayer(m, tLayer, name, opacity=1.0):
    removeAllPopups(m)
    alreadyPresent = False
    tLayer.name = name
    tLayer.opacity = opacity
    for layer in m.layers:
        if name == layer.name:
            m.substitute_layer(layer, tLayer)
            alreadyPresent = True
            break

    if not alreadyPresent:
        m.add_layer(tLayer)

    return tLayer


# Search for a layer of a map by name
def getLayer(m, name):
    for layer in m.layers:
        if name == layer.name:
            return layer
    return None


# Remove a layer given its name
def removeLayer(m, name):
    for layer in m.layers:
        if name == layer.name:
            m.remove_layer(layer)


# Remove all popups from the map
def removeAllPopups(m):
    for layer in reversed(m.layers):
        if isinstance(layer, ipyleaflet.leaflet.Popup):
            m.remove_layer(layer)


# Remove all layer except current basemap
def clear(m):
    baselayer = ipyleaflet.basemap_to_tiles(m.basemap)
    baselayer.base = True
    m.clear_layers()
    m.add_layer(baselayer)


# Zoom a map to a rectangle
def zoomToExtents(m, xmin, ymin, xmax, ymax, epsg=4326):

    m.fit_bounds([[ymin, xmin], [ymax, xmax]])
    return


def zoomToExtents_OLD(m, xmin, ymin, xmax, ymax, epsg=4326):
    def transformPoint(x, y, coordtrans):
        point = ogr.CreateGeometryFromWkt("POINT (%f %f)"%(x,y))
        point.Transform(coordtrans)
        pt = point.GetPoint(0)
        return pt[0], pt[1]

    def transformExtent(xmin, ymin, xmax, ymax, epsgsource, epsgtarget):
        if epsgsource == epsgtarget:
            return xmin, ymin, xmax, ymax

        source = osr.SpatialReference()
        source.ImportFromEPSG(epsgsource)
        source.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

        target = osr.SpatialReference()
        target.ImportFromEPSG(epsgtarget)
        target.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

        ct = osr.CoordinateTransformation(source, target)

        x1, y1 = transformPoint(xmin, ymin, ct)
        x2, y2 = transformPoint(xmin, ymax, ct)
        x3, y3 = transformPoint(xmax, ymin, ct)
        x4, y4 = transformPoint(xmax, ymax, ct)

        targetxmin = min(x1, x2, x3, x4)
        targetxmax = max(x1, x2, x3, x4)
        targetymin = min(y1, y2, y3, y4)
        targetymax = max(y1, y2, y3, y4)

        return targetxmin, targetymin, targetxmax, targetymax

    geoxmin, geoymin, geoxmax, geoymax = transformExtent(xmin, ymin, xmax, ymax, epsg, 4326)
    mercxmin, mercymin, mercxmax, mercymax = transformExtent(geoxmin, geoymin, geoxmax, geoymax, 4326, 3857)

    distance = mercxmax - mercxmin

    # Given a distance on a parallel, returns the number of pixels at a certain zoom level
    def pixels(distance, zoom):
        return 256 * (distance/40000000) * math.pow(2,zoom)

    # See https://gis.stackexchange.com/questions/19632/how-to-calculate-the-optimal-zoom-level-to-display-two-or-more-points-on-a-map
    zoom = 0
    while pixels(distance, zoom) < 500:
        zoom = zoom + 1

    m.center = [(geoymin+geoymax)*0.5, (geoxmin+geoxmax)*0.5]
    m.zoom = zoom


#####################################################################################################################################################
# Display of coordinated at mousemove on a map
#####################################################################################################################################################
# Searches the controls of the Map m to find a control added as a Widget by using its name
def getCardByName(m, name, position='topright', class_='pa-0 ma-0'):
    for control in m.controls:
        if isinstance(control, ipyleaflet.leaflet.WidgetControl):
            if isinstance(control.widget, ipyvuetify.generated.Card):
                if control.widget.active_class == name:
                    return control.widget

    # If not found: add it to the map!
    card = v.Card(flat=True, class_=class_, active_class=name)
    wc = WidgetControl(widget=card, position=position)
    m.add_control(wc)
    return card

    
# Remove a WidgetControl from the map given its name, if present
def removeCardByName(m, name):
    for control in m.controls:
        if isinstance(control, ipyleaflet.leaflet.WidgetControl):
            if isinstance(control.widget, ipyvuetify.generated.Card):
                if control.widget.active_class == name:
                    m.remove_control(control)


#####################################################################################################################################################
# Display of coordinated at mousemove on a map
#####################################################################################################################################################
# Searches the controls of the Map m to find the coordinates control
# N.B.: the Coordinates control is a v.Card having active_class == 'Coordinates' (so that it can be retrieved from the list of m.controls)
def getCoordinatesCard(m):
    return getCardByName(m, 'Coordinates', 'topright')

    
# Remove the Coordinates WidgetControl from the map, if present
def removeCoordinates(m):
    removeCardByName(m, 'Coordinates')


#####################################################################################################################################################
# Functions for vector layers management and display
#####################################################################################################################################################

# Apply a symbol to a category of a VectorLayer instance
def applySymbol(vectorlayer, rule='all', symbol=[]):
    for layer in symbol:
        style = "default"
        for member in layer:
            symbolizer, attribute, value = member
            vectorlayer.set(style, rule, symbolizer, attribute, str(value))


# Generate an image from a symbol
def symbol2Image(symbol=[], size=1, layer_type='Point', clipdimension=MEDIUM_SYMBOLS_DIMENSION, showborder=False):
    doclip = False
    if layer_type == 'Line':
        if size >= 3:    wkt = 'LINESTRING (-170 82, -100 55, -60 70, -10 38)'
        elif size == 2:  wkt = 'LINESTRING (-175 83, -158 81, -148 83, -129 81)'
        else:            wkt = 'LINESTRING (-177 84.45, -171 83.9, -167.4 84.25, -161 83.75)'
    elif layer_type == 'Polygon':
        if size >= 3:    wkt = 'POLYGON ((-170 83.85, -170 10, -10 10, -10 83.85, -170 83.85))'
        elif size == 2:  wkt = 'POLYGON ((-175 84.5, -175 78, -128.5 78, -128.5 84.5, -175 84.5))'
        else:            wkt = 'POLYGON ((-178 84.85, -178 83.2, -160.5 83.2, -160.5 84.85, -178 84.85))'
    else:
        if size >= 3:    wkt = 'POINT (-90 65)'
        elif size == 2:  wkt = 'POINT (-152 82)'
        else:            wkt = 'POINT (-169.52 84.05)'

    if size >= 3:
        if clipdimension < LARGE_SYMBOLS_DIMENSION:
            doclip = True
    elif size == 2:
        if clipdimension < MEDIUM_SYMBOLS_DIMENSION:
            doclip = True
    else:
        if clipdimension < SMALL_SYMBOLS_DIMENSION:
            doclip = True

    v = inter.Collection(inter.collections.Vector)
    v.geomAdd(wkt)
    v.remove('default', 'all')
    applySymbol(v, 'all', symbol)
    ip = v.process()
    ip.toLayer()
    url = SYMBOLS_URL % ip.getProcessID()
    response = requests.get(url)
    if len(response.content) > 5 and response.content[0] == 137 and response.content[1] == 80 and response.content[2] == 78 and response.content[3] == 71 and response.content[4] == 13:
        img = Image.open(BytesIO(response.content))
        if size == 2:
            img = img.crop((0, 0, MEDIUM_SYMBOLS_DIMENSION, MEDIUM_SYMBOLS_DIMENSION))
        elif size < 2:
            img = img.crop((0, 0, SMALL_SYMBOLS_DIMENSION, SMALL_SYMBOLS_DIMENSION))

        if doclip:
            s = img.size
            cx = s[0]/2
            cy = s[1]/2
            img = img.crop((cx-clipdimension/2, cy-clipdimension/2, cx+clipdimension/2, cy+clipdimension/2))
    else:
        # print('URL with errors:',url)
        if size >= 3:    img = Image.new("RGB", (LARGE_SYMBOLS_DIMENSION,  LARGE_SYMBOLS_DIMENSION),  (255, 255, 255))
        elif size == 2:  img = Image.new("RGB", (MEDIUM_SYMBOLS_DIMENSION, MEDIUM_SYMBOLS_DIMENSION), (255, 255, 255))
        else:            img = Image.new("RGB", (SMALL_SYMBOLS_DIMENSION,  SMALL_SYMBOLS_DIMENSION),  (255, 255, 255))
        draw = ImageDraw.Draw(img)
        draw.text((0, 0),"Error",(0,0,0))

    # Add a thin black border
    if showborder:
        draw = ImageDraw.Draw(img)
        s = img.size
        draw.rectangle(((0, 0), (s[0]-1, s[1]-1)), outline='black')
    return img


# Change color of a symbol and returns the modified symbol
def setColor(symbol, color='#ff0000', fillColor='#ff0000', fillOpacity=1.0, strokeColor='#ffff00', strokeWidth=0.5, scalemin=None, scalemax=None):
    newsymbol = []
    for layer in symbol:
        newlayer = []
        for member in layer:
            symbolizer,attribute,value = member

            if value == 'COLOR':
                value = color

            if value == 'FILL-COLOR':
                value = fillColor

            if value == 'FILL-OPACITY':
                value = fillOpacity

            if value == 'STROKE-COLOR':
                value = strokeColor

            if value == 'STROKE-WIDTH':
                value = strokeWidth

            if value == 'SCALE-MIN':
                value = scalemin

            if value == 'SCALE-MAX':
                value = scalemax

            if value is not None:
                newlayer.append((symbolizer,attribute,value))

        newsymbol.append(newlayer)

    return newsymbol


#####################################################################################################################################################
# Save a map view in a PIL image
#####################################################################################################################################################
def toImage(m):

    # Bounds and zoom of the current view
    (latmin, lonmin), (latmax, lonmax) = m.bounds
    zoom = m.zoom

    # URLs of all the Tilelayer on the map
    baseUrls = [x.url for x in m.layers if type(x) == ipyleaflet.leaflet.TileLayer]

    # Opacities
    opacities = [x.opacity for x in m.layers if type(x) == ipyleaflet.leaflet.TileLayer]

    # Convert lat/lon/zoom to xtile,ytile.
    # See https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
    def latlon2tile(lat_deg, lon_deg, zoom):
        lat_rad = (lat_deg * math.pi) / 180.0
        n = math.pow(2,zoom)
        xtile = n * ((lon_deg + 180.0) / 360.0)
        ytile = n * (1 - (math.log(math.tan(lat_rad) + 1.0/math.cos(lat_rad)) / math.pi)) / 2
        return xtile, ytile

    xtile1f, ytile2f = latlon2tile(latmin, lonmin, zoom)
    xtile2f, ytile1f = latlon2tile(latmax, lonmax, zoom)

    xtile1 = int(xtile1f)
    xtile2 = int(xtile2f)
    ytile1 = int(ytile1f)
    ytile2 = int(ytile2f)

    # Amount of pixels to crop on each side
    dx1 = 256*(xtile1f-xtile1)
    dx2 = 256*(xtile2+1-xtile2f)
    dy1 = 256*(ytile1f-ytile1)
    dy2 = 256*(ytile2+1-ytile2f)

    dx1 = round(dx1*100)//100
    dx2 = round(dx2*100)//100
    dy1 = round(dy1*100)//100
    dy2 = round(dy2*100)//100

    # Number of tiles
    nx = xtile2 - xtile1 + 1
    ny = ytile2 - ytile1 + 1

    # Dimension of the overall image
    w = 256 * nx
    h = 256 * ny
    imageTotal = Image.new(mode="RGBA", size=(w,h))

    # Substitute x,y,z into a TileService URL
    def url(baseurl, x, y, zoom):
        return baseurl.replace('{x}', str(int(x))).replace('{y}', str(int(y))).replace('{z}', str(int(zoom)))

    # Cycle on all tiles and compose the overall image
    for x in range(nx):
        xt = xtile1 + x
        xpos = x*256
        for y in range(ny):
            yt = ytile1 + y
            ypos = y*256
            for baseurl, opacity in zip(baseUrls, opacities):
                image = Image.open(requests.get(url(baseurl, xt, yt, zoom), stream=True).raw)
                image = image.convert('RGBA')

                if opacity < 1.0:
                    # Split image in 4 channels
                    (r, g, b, a) = image.split()

                    # Change the alpha channel
                    # See https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.eval
                    a = Image.eval(a, lambda px: opacity*px)

                    # Merge 4 channels
                    image = Image.merge('RGBA', (r, g, b, a))

                # Transparent paste!!!
                # See https://stackoverflow.com/questions/5324647/how-to-merge-a-transparent-png-image-with-another-image-using-pil
                imageTotal.paste(image, (xpos, ypos), mask=image)

    # Crop the image
    area_crop = (dx1, dy1, w-dx2, h-dy2)
    return imageTotal.crop(area_crop)


"""Widget to select among alternative display using a list of tabs displayed horizontally or vertically."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import tooltip
except:
    import settings
    import tooltip


#####################################################################################################################################################
# Tabs control
#####################################################################################################################################################
class tabs:
    """
    Widget to select among alternative display using a list of tabs displayed horizontally or vertically.
        
    Parameters
    ----------
    index : int
        Index of the selected option at start (from 0 to len(labels)-1)
    labels : list of strings
        Strings to be displayed as text of the options
    contents : list of widgets, optional
        Widgets to be alternatively displayed when each of the tabs option is selected (for instance could be a list of ipywidgets.Output widgets). Default is None
    tooltips : list of strings, optional
        List of strings to be used as tooltips for the single tabs, in order (default is None)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    dark : bool, optional
        Flag to invert the text and backcolor (default is the value of settings.dark_mode)
    onchange : function, optional
        Python function to call when the user clicks on one of the tabs. The function will receive a parameter of type int containing the index of the selected tab, from 0 to len(labels)-1
    row : bool, optional
        Flag to display the tabs horizontally or vertically (default is True)
            
    Example
    -------
    Creation and display of a tabs widget to select among alternative Outputs display::
        
        from vois.vuetify import tabs
        from ipywidgets import widgets
        from IPython.display import display

        debug = widgets.Output()

        output0 = widgets.Output()
        output1 = widgets.Output()
        output2 = widgets.Output()

        with output0: print('This is output 0')
        with output1: print('This is output 1')
        with output2: print('This is output 2')

        def onchange(index):
            with debug:
                print(index)

        t = tabs.tabs(0, ['Option 0', 'Option 1', 'Option 2'],
                      contents=[output0,output1,output2],
                      onchange=onchange, row=False)

        display(t.draw())
        display(debug)

    .. figure:: figures/tabs.png
       :scale: 100 %
       :alt: tabs widget

       Creation of a tabs widget to display ipywidgets.Output content at every selection.
   """

    # Initialization
    def __init__(self, index, labels, contents=None, tooltips=None, color=settings.color_first, dark=settings.dark_mode, onchange=None, row=True):
        
        self.index    = index
        self.labels   = labels
        self.onchange = onchange
        
        s = ''
        if dark: s = "color: %s;" % settings.textcolor_dark
            
        self.tab_list = []
        self.tab_list_with_tooltips = []
        i = 0
        for index,label in enumerate(self.labels):
            t = v.Tab(children=[label], style_=s, disabled=False)
            t.on_event('click', self.__internal_onchange)
            if isinstance(tooltips, list) and len(tooltips) > index:
                self.tab_list_with_tooltips.append(tooltip.tooltip(tooltips[index],t))
            else:
                self.tab_list_with_tooltips.append(t)
                
            self.tab_list.append(t)
            i += 1
            
        content_list = []
        if not contents is None:
            content_list = [v.TabItem(children=[x]) for x in contents]
            
        slider = v.TabsSlider(color=color, dark=dark)
        
        backcolor = 'white'
        textcolor = settings.textcolor_notdark
        if dark:
            backcolor = settings.dark_background
            textcolor = settings.textcolor_dark

        self.tabswidget = v.Tabs(v_model=self.index, vertical=not row, dense=True, class_='pa-0 ma-0', 
                                 #background_color='white', color=settings.textcolor_notdark,
                                 background_color=backcolor, color=textcolor, 
                                 children=[slider] + self.tab_list_with_tooltips + content_list)
        
   

    # Manage onchange event
    def __internal_onchange(self, widget, event, data):
        self.index = self.tab_list.index(widget)
        if self.onchange:
            self.onchange(self.index)
    
    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Tabs widget)"""
        return self.tabswidget

    
    # Get the active tab
    @property
    def value(self):
        """
        Get/Set the active tab index.
        
        Returns
        --------
        index : int
            Index of the selected tab (from 0 to len(labels)-1)

        Example
        -------
        Programmatically select one of the tab and print the index of the selected option::
            
            t.value = 2
            print(t.value)
        
        """
        return self.index
   
    
    # Set the active button
    @value.setter
    def value(self, index):
        if index >= 0 and index < len(self.labels):
            self.index = index
            self.tabswidget.v_model = self.index
            if self.onchange:
                self.onchange(self.index)

                
    # disabled property
    @property
    def disabled(self):
        """
        Get/Set the disabled state.
        """
        return max([x.disabled for x in self.tab_list])
    
    @disabled.setter
    def disabled(self, flag):
        for tab in self.tab_list:
            tab.disabled = flag


"""Simplified creation of v-treeview vuetify widget to display hierarchical data in a tree"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import traitlets
import ipyvuetify as v
from ipywidgets import widgets, Layout, HTML
from IPython.display import display
import pandas as pd
import copy

try:
    from . import settings
except:
    import settings


##################################################################################################################################
# Creation of ipyvuetify Treeview
##################################################################################################################################

##################################################################################################################################
# Customized Treeview created using VuetifyTemplate
# The nodes of the tree can have a checkbox (selectable=True) and/or can be activated (activatable=True)
# The height of each node is reduced, the font is smaller and the paddings/margins are reduced
# See last cell of https://github.com/mariobuikhuizen/ipyvuetify/blob/master/examples/Examples.ipynb
##################################################################################################################################
class CustomTreeview(v.VuetifyTemplate):
    """
    Ipyvuetify template to display a custom treeview. The nodes of the tree can have a checkbox (selectable=True) and/or can be activated (activatable=True).

    Attributes
    ----------
    items : list, optional
        JSON object to represent the tree: each object has 'id' and 'name' key, 'disabled', 'isfolder' and 'icon' are optional keys (default is [])
    selectable: bool, optional
        If True the nodes of the tree have a checkbox to select them (default is True)
    activatable: bool, optional
        If True, one of the nodes of the tree can be activated (default is False)
    selected: list, optional
        List of id of the nodes that are selected on start (default is [])
    selectednames: list, optional
        List of name of the nodes that are selected at any time (default is [])
    opened: list, optional
        List of id of the nodes that are to be opened on start (default is [])
    color: str, optional
        Color for the selected nodes (default is 'blue')
    on_change: function, optional
        Python function to call when the selection of the tree items changes (default is None)
    on_activated: function, optional
        Python function to call when the active item changes (user selecting a node) (default is None)
    expand_selection_to_parents: bool, optional
        If True, also the parent nodes are returned as selected when all children are selected (default is True)
    iconsshow : bool, optional
        If True, an icon is added to each node of the tree (default is False)
    iconscolor : str, optional
        Color of the icons (default is 'blue')
    icons_folder_opened : str, optional
        Name of the icon to use for opened nodes that have children when iconsfolder is True (default 'mdi-folder-open')
    icons_folder_closed : str, optional
        Name of the icon to use for closed nodes that have children when iconsfolder is True (default 'mdi-folder')
    tooltips : bool, optional
        If True the nodes will show the tooltip (default is False)
    tooltips_chars : int, optional
        Minimum lenght of the node label to show the tooltip (default is 20)
    search : str, optional
        Filter to display only the nodes of the tree that contain the text (default is '' which means that all the nodes of the tree are displayed)
    item_height : int, optional
        Item height in pixels (default is 24)
    font_size : int, optional
        Font size in pixels (default is 15)
    icon_size : int, optional
        Icon size in pixels (default is 18)
    checkbox_size : int, optional
        Checkbox size in pixels (default is 24)
        
    Note
    ----
    This class is not intended to be called directly, but only through the functions :func:`~treeview.createTreeviewFromList` and :func:`~treeview.createTreeviewFromDF2Columns`.
    """
    
    items                       = traitlets.Any([]).tag(sync=True)                       # JSON object to represent the tree (each object has 'id' and 'name' key. 'disabled' is an optional key)
    selectable                  = traitlets.Bool(True).tag(sync=True)                    # If True the nodes of the tree have a checkbox to select them
    activatable                 = traitlets.Bool(False).tag(sync=True)                   # If True, one the nodes of the tree can be activated
    open_on_click               = traitlets.Bool(False).tag(sync=True)                   # If True, the nodes can be opened by clicking on the label
    active                      = traitlets.Any(None).tag(sync=True)                     # Id of the node that is active at start
    selected                    = traitlets.Any([]).tag(sync=True)                       # List of id of the nodes that are selected on start
    selectednames               = traitlets.Any([]).tag(sync=True)                       # List of name of the nodes that are selected at any time
    opened                      = traitlets.Any([]).tag(sync=True)                       # List of id of the nodes that are to be opened on start
    color                       = traitlets.Unicode('blue').tag(sync=True)               # Color for the selected nodes
    dark                        = traitlets.Bool(False).tag(sync=True)                   # If True show text in white
    transition                  = traitlets.Bool(False).tag(sync=True)                   # If True applies a transition when nodes are opened and closed
    on_change                   = traitlets.Any(None).tag(sync=False)                    # Name of a python function to call when the selected nodes change (user clicking in one of the checkboxes)
    on_activated                = traitlets.Any(None).tag(sync=False)                    # Name of a python function to call when the active node changes (user selecting a node)
    expand_selection_to_parents = traitlets.Bool(True).tag(sync=True)                    # If True, also the parent nodes are returned as selected when all children are selected
    search                      = traitlets.Unicode('').tag(sync=True)                   # Search string to filter the nodes
    opened_all                  = traitlets.Bool(False).tag(sync=True)                   # If True all the nodes are opened at start
    
    
    @traitlets.default('template')
    def _template(self):

        # Code to add to activate the icons
        # if item.isfolder is True, the open/closed icon is added, else the item.icon is added
        icons_template = ''
        if self.iconsshow:
            icons_template = '''
<template v-slot:prepend="{ item, open }">
      <v-icon color="%s" :disabled="item.disabled" v-if="item.isfolder">
         {{ open ? '%s' : '%s' }}
      </v-icon>
      <v-icon color="%s" :disabled="item.disabled" v-else-if="item.icon">
         {{ item.icon }}
      </v-icon>
      <v-html v-else>
      </v-html>
</template>
''' % (self.iconscolor, self.icons_folder_opened, self.icons_folder_closed, self.iconscolor)


        # Code to add to activate the tooltip
        tooltip_template = ''
        if self.tooltips:
            tooltip_template = '''
<template v-slot:label="{ item }">
    <v-tooltip bottom :disabled="item.name.length < %d">
      <template v-slot:activator="{ on, attrs }">
        <span v-bind="attrs" v-on="on">
           {{ item.name }}
        </span>
      </template>
      {{ item.name }}
    </v-tooltip>
</template>
''' % self.tooltips_chars

            
        return '''
<v-treeview 
    :selectable="selectable"
    :activatable="activatable"
    :active="[active]"
    hoverable
    dense
    :dark="dark"
    :transition="transition"
    class="pa-0 ma-0"
    :items="items"
    v-model="selected"
    :search="search"
    :open="opened"
    :open-all="opened_all"
    :color="color"
    :selected-color="color"
    :open-on-click="open_on_click"
    @input="change_selection"
    @update:active="activate"
    @update:open="onopening">
    %s
    %s
</v-treeview>

<style id="treeview-item-style">

.vuetify-styles .v-treeview-node__root {
    padding-left: 0px;
    padding-right: 0px;
}
.vuetify-styles .v-treeview--dense .v-treeview-node__root {
    min-height: %dpx;
    font-size: %dpx;
}
.vuetify-styles .v-application--is-ltr .v-treeview-node__checkbox {
    margin-left: 0px;
    margin-right: 0px;
}

.vuetify-styles .v-treeview-node__checkbox {
    width: %dpx;
}
.vuetify-styles .v-icon.v-icon {
    font-size: %dpx;
}

</style>
''' % (tooltip_template, icons_template, self.item_height, self.font_size, self.checkbox_size, self.icon_size)
    
    
    # Expand the selection to parents that have all childrens selected
    def expandSelectionToParents(self, node):
        if 'children' in node:
            sel = True
            for n in node['children']:
                self.expandSelectionToParents(n)
                if n['id'] not in self.selected: sel = False
            if sel and not node['id'] in self.selected:
                self.selected.append(node['id'])
    
    
    # Transform a list of id in a list of fullnames
    def id2Names_selected(self, node, idlist):
        if node['id'] in idlist: self.selectednames.add(node['fullname'])
        if 'children' in node:
            for n in node['children']:
                self.id2Names_selected(n,idlist)

    # Return the fullname given the id
    def id2Name(self, node, id):
        if node['id'] == id:
            return node['fullname']
        if 'children' in node:
            for n in node['children']:
                res = self.id2Name(n,id)
                if not res is None:
                    return res

                
    # Updates self.selectednames with the selected nodes
    def updateSelectedNames(self):
        if self.expand_selection_to_parents:
            for i in self.items:
                self.expandSelectionToParents(i)
        self.selected.sort()
        self.selectednames = set()
        for i in self.items:
            self.id2Names_selected(i,self.selected)
        self.selectednames = sorted(list(self.selectednames))
    
    # Manage event "input": when a checkbox of the tree is clicked
    def vue_change_selection(self, data):
        #print(data)
        self.selected = data
        if self.selectable and not self.on_change is None:
            self.updateSelectedNames()
            self.on_change(self.selectednames)
            
    # Manage the activation (selection) of a node of the tree
    def vue_activate(self, data):
        if len(data) > 0:
            self.current_active = data[0]
            if self.activatable and not self.on_activated is None:
                for i in self.items:
                    name = self.id2Name(i,data[0])
                    if not name is None:
                        self.on_activated(name)
                        break
        else:
            self.current_active = None
            if self.activatable and not self.on_activated is None:
                self.on_activated(None)

    # Called every time a node is opened or closed
    def vue_onopening(self, data):
        self.current_open = data
                
            
    # Initialization
    def __init__(self,
                 items                       = [],
                 selectable                  = True,
                 activatable                 = False,
                 open_on_click               = False,
                 active                      = None,
                 selected                    = [],
                 selectednames               = [],
                 opened                      = [],
                 color                       = 'blue',
                 dark                        = False,
                 transition                  = False,
                 on_change                   = None,
                 on_activated                = None,
                 expand_selection_to_parents = True,
                 search                      = '',
                 opened_all                  = False,
                 item_height                 = 24,     # Item height in pixels
                 font_size                   = 15,     # Font size in pixels
                 icon_size                   = 18,     # Icon size in pixels
                 checkbox_size               = 24,     # Checkbox size in pixels
                 
                 iconsshow=False,
                 iconscolor='blue',
                 icons_folder_opened='mdi-folder-open',
                 icons_folder_closed='mdi-folder',
                 tooltips=False,
                 tooltips_chars=20,
                 *args,
                 **kwargs):
        
        self.items                       = items
        self.selectable                  = selectable
        self.activatable                 = activatable
        self.open_on_click               = open_on_click
        self.active                      = active
        self.selected                    = selected
        self.selectednames               = selectednames
        self.opened                      = opened
        self.color                       = color
        self.dark                        = dark
        self.transition                  = transition
        self.on_change                   = on_change
        self.on_activated                = on_activated
        self.expand_selection_to_parents = expand_selection_to_parents
        self.search                      = search
        self.opened_all                  = opened_all
        
        self.item_height                 = item_height      # Item height in pixels
        self.font_size                   = font_size        # Font size in pixels
        self.icon_size                   = icon_size        # Icon size in pixels
        self.checkbox_size               = checkbox_size    # Checkbox size in pixels
        
        self.current_active = active
        self.current_open   = []
        
        self.iconsshow = iconsshow
        self.iconscolor = iconscolor
        self.icons_folder_opened = icons_folder_opened
        self.icons_folder_closed =icons_folder_closed
        self.tooltips = tooltips
        self.tooltips_chars = tooltips_chars
        super().__init__(*args, **kwargs)
        
        

##################################################################################################################################
# Helper class to operate on a treeview returned by the createTreeviewFromList or createTreeviewFromDF2Columns functions
##################################################################################################################################
class treeviewOperations():
    """
    Helper class to operate on a treeview returned by the :func:`~treeview.createTreeviewFromList` and :func:`~treeview.createTreeviewFromDF2Columns` functions. 
    This class allows for activation and opening/closing of nodes of the tree
    
    Attributes
    ----------
    treecard : instance of ipyvuetify Card widget
        Value returned by a call to the functions :func:`~treeview.createTreeviewFromList` and :func:`~treeview.createTreeviewFromDF2Columns`.
        
    
    Example
    -------
    Creation and display of a treeview and programmatical activation and opening of the nodes::
        
        from vois.vuetify import treeview
        from IPython.display import display
        
        treecard = treeview.createTreeviewFromList(['A','A.1','A.2','A.1.1',
                                                    'A.3.1','A.4.1.2','A.5.2.1',
                                                    'A.4.2.3.1','B','B.A'],
                                                   rootName='Root', 
                                                   activatable=True,
                                                   expand_selection_to_parents=False, 
                                                   substitutionDict={'A.1': 'A.1 new name'},
                                                   color='green',
                                                   height=350)
        display(treecard)
        
        top = treeview.treeviewOperations(treecard)
        
        # Print active node
        print(top.getActive())
        
        # Set active node
        top.setActive('A.1.1')
        
        # Set the list of opened nodes
        top.setOpened(['Root', 'A', 'A.3'])
        
    """
    
    # Initialisation
    def __init__(self, treecard):
        
        # Retrieve the underlying instance of CustomTreeview class
        self.treeview = treecard.children[0].children[0]
        
        self.doindex()

    
    # calculate indexing of the treeview nodes
    def doindex(self):
        # Dictionaries to index the three nodes
        # id2fullname    Map nodes id to the fullname                 # key: id           value : fullname of the node
        # fullname2id    Map nodes fullname to the id                 # key: fullname     value : id of the node
        # id2parentid    Map node id to the id of the parent node     # key: id           value : id of the parent node
        # id2childrenid  Map node id to the list of children ids      # key: id           value : list of id of the children nodes
        self.id2fullname   = {}
        self.fullname2id   = {}
        self.id2parentid   = {}
        self.id2childrenid = {}
        for i in self.treeview.items:
            self.__indexItems(i)
        
        
    # Internal recursive function to create the indexing of the treeview nodes
    def __indexItems(self,node):
        self.id2fullname[node['id']]       = node['fullname']
        self.fullname2id[node['fullname']] = node['id']
        if 'children' in node:
            self.id2childrenid[node['id']] = []
            for n in node['children']:
                self.id2parentid[n['id']] = node['id']
                self.id2childrenid[node['id']].append(n['id'])
                self.__indexItems(n)
        

    # Get the fullnames of the opened nodes
    def getOpened(self):
        """
        Returns the list of the fullnames of the opened nodes of the treeview
        """
        #fullnames = [self.id2fullname[x] for x in self.treeview.opened]
        fullnames = [self.id2fullname[x] for x in self.treeview.current_open]
        return fullnames

    
    # Set the nodes opened given a list of fullnames of the nodes
    def setOpened(self, fullnames):
        """
        Set the list of opened nodes of the treeview given their fullnames
        """
        ids = list(set([self.fullname2id[x] for x in fullnames]))
        self.treeview.opened = ids

        
    # Get the fullnames of the selected nodes
    def getSelected(self):
        """
        Returns the list of the fullnames of the selected nodes of the treeview
        """
        fullnames = [self.id2fullname[x] for x in self.treeview.selected]
        return fullnames
        
        
    # Set the nodes selected given a list of fullnames of the nodes
    def setSelected(self, fullnames):
        """
        Set the list of selected nodes of the treeview given their fullnames
        """
        ids = list(set([self.fullname2id[x] for x in fullnames]))
        self.treeview.selected = ids
        
        
    # Open all the nodes of the tree
    def openAll(self):
        """
        Open all the nodes of the treeview
        """
        allfullnames = list(self.fullname2id.keys())
        self.setOpened(allfullnames)

        
    # Get the fullname of the active node
    def getActive(self):
        """
        Returns the fullname of the node that is active in the treeview
        """
        nodeid = self.treeview.current_active
        if not nodeid is None:
            return self.id2fullname[nodeid]
        return None

    
    # Set the active node by passing its fullname
    def setActive(self, fullname):
        """
        Set the active node of the treeview by passing its fullname
        """
        if fullname in self.fullname2id:
            nodeid = self.fullname2id[fullname]
            self.treeview.active = nodeid
            self.treeview.current_active = nodeid
            
            # Opens all the parent nodes
            idparents = []
            while not nodeid is None:
                if nodeid in self.id2parentid:
                    nodeid = self.id2parentid[nodeid]
                    if not nodeid is None:
                        idparents.append(nodeid)
                else:
                    nodeid = None
            if not 1 in idparents:
                idparents.append(1)
            
            op = set(self.treeview.opened + idparents)
            self.treeview.opened = list(op)
            
            
    # Given the fullname of a node, returns the fullname of its first child, or None if the node has no children
    def getFirstChildFullname(self, fullname):
        """
        Given the fullname of a node, returns the fullname of its first child, or None if the node has no children
        """
        if fullname in self.fullname2id:
            nodeid = self.fullname2id[fullname]
            if nodeid in self.id2childrenid:
                return self.id2fullname[self.id2childrenid[nodeid][0]]
        return None
    
    
    # Search for a text in the nodes of the tree
    def setSearch(self, text2search=''):
        """
        Search for a text string in the nodes of the tree
        """
        self.treeview.search = text2search

        
    # Retrieve the children of a node
    def getChildren(self, nodefullname):
        """
        Returns the list of full names of the children of a node
        """
        
        # Recursive function
        def __getchildren(node, nodeid):
            if node['id'] == nodeid:
                if 'children' in node:
                    return [x['id'] for x in node['children']]

            if 'children' in node:
                for n in node['children']:
                    res = __getchildren(n, nodeid)
                    if len(res) > 0:
                        return res
            return []
        
        if nodefullname in self.fullname2id:
            nodeid = self.fullname2id[nodefullname]
            res = __getchildren(self.treeview.items[0],nodeid)
            return [self.id2fullname[x] for x in res]
        
        return []
        
        
        
    # Dynamic setting of children to nodes
    def setChildren(self, nodefullname, childrenfullnames):
        """
        Dynamically change the children of a node

        Example
        -------
        Creation and display of a treeview and programmatically add children to nodes::
        
            from vois.vuetify import treeview
            from IPython.display import display

            treecard = treeview.createTreeviewFromList(['A','A.1','A.2','B'],
                                                       separator='.',
                                                       rootName='Root',
                                                       expand_selection_to_parents=False, 
                                                       color='green',
                                                       width='500px',
                                                       height='350px',
                                                       selectable=False,
                                                       activatable=True)

            display(treecard)
            
            top = treeview.treeviewOperations(treecard)
            
            top.setChildren('A.2', ['A.2.1', 'A.2.2'])

        """
        
        # Recursive function
        def __setchildren(node, nodeid, children):
            if node['id'] == nodeid:
                node['children'] = children

            if 'children' in node:
                for n in node['children']:
                    __setchildren(n, nodeid, children)
        
        if nodefullname in self.fullname2id:
            nodeid = self.fullname2id[nodefullname]
            nextid = 1 + max(self.id2fullname.keys())
            children = []
            for name in childrenfullnames:
                c = {'id': nextid, 'name': name, 'fullname': name}
                children.append(c)
                nextid += 1
            
            root = copy.deepcopy(self.treeview.items[0])
            __setchildren(root, nodeid, children)
            self.treeview.items = [root]
            self.doindex()
    
    
##################################################################################################################################
# Create a ipyvuetify Treeview from a list of names with an implicit tree structure, example: ['JRC', 'JRC.D', 'JRC.D.3', ...]
# Returns an ipyvuetify widget containing the tree
# Example:
# createTreeviewFromList(['A','A.1','A.2','A.1.1'], rootName='A')
##################################################################################################################################
def createTreeviewFromList(nameslist=[],
                           rootName='Root',
                           separator='.', 
                           select_all=True,
                           on_change=None,
                           on_activated=None,
                           expand_selection_to_parents=True,
                           displayfullname=True,
                           width='200px', height='500px',
                           elevation=0,
                           repeat_parent_as_first_child=False,
                           substitutionDict=None,
                           color=settings.color_first,
                           dark=settings.dark_mode,
                           transition=False,
                           selectable=True,
                           activatable=False,
                           active=None,
                           open_on_click=False,
                           opened=[],
                           opened_all=False,
                           selected=[],
                           disabled=[],
                           iconsshow=False,
                           iconscolor=settings.color_first,
                           iconsfolder=True,
                           icons_folder_opened='mdi-folder-open',
                           icons_folder_closed='mdi-folder',
                           iconroot=None,
                           iconsDict=None,
                           tooltips=False,
                           tooltips_chars=20,
                           item_height=24,
                           font_size=15,
                           icon_size=18,
                           checkbox_size=24):
    """
    Create a treeview form a list of strings and a separator that defines the hierarchical structure (example: ['A', 'A.1', 'A.2', 'B', 'B.3']).

    Parameters
    ----------
    nameslist : list, optional
        List of strings that contain a hierarchical structure, considering the separator character (default is [])
    rootName : str, optional
        Name to be displayed as root of the tree (default is 'Root')
    separator : str, optional
        String or character to be considered as separator for extracting the hierarchical structure from the nameslist list of strings (default is '.')
    select_all : bool, optional
        Flag to control the initial selection of all the nodes of the tree (default is True)
    on_change : function, optional
        Python function to call when the selected nodes change caused by user clicking in one of the checkboxes (default is None)
    on_activated : function, optional
        Python function to call when the active item changes (user selecting a node) (default is None)
    expand_selection_to_parents : bool, optional
        If True, also the parent nodes are returned as selected when all children are selected (default is True)
    displayfullname : bool, optional
        If True the nodes will display the full names, id False only the last part splitted by the separator (default is True)
    width : str, optional
        Width of the treeview widget (default is '200px')
    height : str, optional
        Height of the treeview widget (default is '500px')
    elevation : int, optional
        Elevation to assign to the widget (default is 0)
    repeat_parent_as_first_child : bool, optional
        If True each parent node will have a first children with its name (default is False)
    substitutionDict: dict, optional
        Dictionary to apply substitutions to the fullnames of the items extracted from the nameslist parameters (default is None)
    color : str, optional
        Color to be used as main color of the Treeview widget (default is settings.color_first)
    dark : bool, optional
        If True, the treeview widget will have a dark background (default is settings.dark_mode)
    transition : bool, optional
        If True applies a transition when nodes are opened and closed (default is False)
    selectable : bool, optional
        If True the nodes of the tree have a checkbox to select them (default is True)
    activatable : bool, optional
        If True, one of the nodes of the tree can be activated (default is False)
    active : str, optional
        Name of the node to activate on start (default is None)
    open_on_click : bool, optional
        If True, the nodes of the tree can be opened also by clicking on the node label (default is False)
    opened : list of str, optional
        List of names of the nodes to open at start (default is [])
    opened_all : bool, optional
        If True, all the tree nodes are opened at start (default is False). This setting has prevalence over the opened parameter.
    selected : list of str, optional
        List of fullnames of the nodes to select at start (default is []).
    disabled : list of str, optional
        List of fullnames of the nodes to display as disabled in the tree (default is [])
    iconsshow : bool, optional
        If True, an icon is added to each node of the tree (default is False)
    iconscolor : str, optional
        Color of the icons (default is settings.color_first)
    iconsfolder : bool, optional
        If True (and if iconsshow is True) it adds the open/closed folder icon to nodes that have children (default is True)
    icons_folder_opened : str, optional
        Name of the icon to use for opened nodes that have children when iconsfolder is True (default 'mdi-folder-open')
    icons_folder_closed : str, optional
        Name of the icon to use for closed nodes that have children when iconsfolder is True (default 'mdi-folder')
    iconroot : str, optional
        Name of the icon for the root node of the tree (default is None)
    iconsDict: dict, optional
        Dictionary to apply icons to the fullnames of the items, if iconsshow is True (default is None)
    tooltips : bool, optional
        If True the nodes will show the tooltip (default is False)
    tooltips_chars : int, optional
        Minimum lenght of the node label to show the tooltip (default is 20)
    item_height : int, optional
        Item height in pixels (default is 24)
    font_size : int, optional
        Font size in pixels (default is 15)
    icon_size : int, optional
        Icon size in pixels (default is 18)
    checkbox_size : int, optional
        Checkbox size in pixels (default is 24)

    Returns
    -------
    v.Card: ipyvuetify Card widget
        An ipyvuetify v.Card widget having a v.Html as its only child. The v.Html has a treeview.CustomTreeview widget as its only child

    Example
    -------
    Creation and display of a treeview::
    
        from vois.vuetify import treeview
        from IPython.display import display
        
        treecard = treeview.createTreeviewFromList(['A','A.1','A.2','A.1.1',
                                                    'A.3.1','A.4.1.2','A.5.2.1',
                                                    'A.4.2.3.1','B','B.A'], 
                                                   rootName='Root', 
                                                   expand_selection_to_parents=False, 
                                                   substitutionDict={'A.1': 'A.1 new name'},
                                                   color='green',
                                                   height='350px')
        display(treecard)

    .. figure:: figures/treeview1.png
       :scale: 100 %
       :alt: treeview widget

       Treeview widget created from a list of strings.
    """
    
    # Return the string to display in the tree for a node, given its fullname
    def GetNodeString(fullname):
        if displayfullname:
            return fullname
        else:
            parts = fullname.split(separator)
            if len(parts) > 0:
                return parts[-1]
            return fullname
        

    # Add a new node to the tree
    def addNode(name, parent):
        nonlocal nextid, nodes, names2id
        if not name in nodes:
            elemname = name
            if not substitutionDict is None:
                if elemname in substitutionDict:
                    elemname = substitutionDict[elemname]
            elem = {'id': nextid, 'name': GetNodeString(elemname), 'fullname': elemname }
            if name in disabled: elem['disabled'] = True
            if not iconsDict is None:
                if elemname in iconsDict:
                    elem['icon'] = iconsDict[elemname]
            names2id[elemname] = nextid
            nextid += 1
            nodes[name] = elem
            if parent != None:
                if not 'children' in parent:
                    parent['children'] = []
                    
                    if repeat_parent_as_first_child and parent['name'] != rootName:
                        firstelem = {'id': nextid, 'name': GetNodeString(parent['name']), 'fullname': parent['name'] }
                        if parent['name'] in disabled: firstelem['disabled'] = True
                        names2id[parent['name']] = nextid
                        nextid += 1
                        parent['children'].append(firstelem)
                    
                parent['children'].append(elem)
                if iconsshow and iconsfolder: parent['isfolder'] = True
                
            return elem
    
    nextid = 1
    nodes  = {}
    names2id = {}  # Convert a fullname to its id

    root = addNode(rootName,None)

    # Manage an implicit tree with separator!
    for name in nameslist:
        parts = name.split(separator)
        fullname = ''
        for part in parts:
            parentname = fullname
            if len(fullname) == 0: fullname += part
            else:                  fullname += separator + part

            parent = root
            if parentname in nodes: parent = nodes[parentname]

            addNode(fullname,parent)

    # Returns the list of all the ids of the children of node
    def calculateIdsOfLeaves(node, listofids):
        if not 'children' in node:
            listofids.append(node['id'])
        else:
            for n in node['children']:
                listofids = calculateIdsOfLeaves(n,listofids)
        return listofids
    
    selectedids = []
    if select_all: selectedids = calculateIdsOfLeaves(root,[])
    else:          selectedids = [names2id[x] for x in selected if x in names2id]
    
    
    # Convert opened names to ids
    openedids = []
    if not opened_all:
        if len(opened) == 0: openedids = [1]  # Only Root is opened by default
        else:                openedids = [names2id[x] for x in opened if x in names2id]

    # Convert active name to id
    if not active is None:
        activeid = names2id[active]
    else:
        activeid = None
    
    tree = CustomTreeview(items=[root], selected=selectedids, opened=openedids, on_change=on_change, on_activated=on_activated, 
                          expand_selection_to_parents=expand_selection_to_parents, dark=dark, transition=transition, opened_all=opened_all, 
                          color=color, selectable=selectable, activatable=activatable, active=activeid, open_on_click=open_on_click,
                          tooltips=tooltips, tooltips_chars=tooltips_chars,
                          iconsshow=iconsshow, iconscolor=iconscolor, icons_folder_opened=icons_folder_opened, icons_folder_closed=icons_folder_closed,
                          item_height=item_height, font_size=font_size, icon_size=icon_size, checkbox_size=checkbox_size)
    treehtml = v.Html(tag='div', height=height, children=[tree], style_='overflow: hidden;')
    treecard = v.Card(width=width, height=height, elevation=elevation, children=[treehtml],
                      dark=dark, style_='overflow-x: visible;')
    
    return treecard




##################################################################################################################################
# Create a flat ipyvuetify Treeview from a list of names
# Returns an ipyvuetify widget containing the tree
# Example:
# createFlatTreeview(['A','B','C','D'])
##################################################################################################################################
def createFlatTreeview(nameslist=[],
                       select_all=True,
                       on_change=None,
                       on_activated=None,
                       displayfullname=True,
                       width='200px', height='500px',
                       elevation=0,
                       color=settings.color_first,
                       dark=settings.dark_mode,
                       transition=False,
                       selectable=True,
                       activatable=False,
                       active=None,
                       selected=[],
                       disabled=[],
                       iconsshow=False,
                       iconscolor=settings.color_first,
                       iconsDict=None,
                       tooltips=False,
                       tooltips_chars=20,
                       item_height=24,
                       font_size=15,
                       icon_size=18,
                       checkbox_size=24):
    """
    Create a flat treeview form a list of strings

    Parameters
    ----------
    nameslist : list, optional
        List of strings that contain a hierarchical structure, considering the separator character (default is [])
    select_all : bool, optional
        Flag to control the initial selection of all the nodes of the tree (default is True)
    on_change : function, optional
        Python function to call when the selected nodes change caused by user clicking in one of the checkboxes (default is None)
    on_activated : function, optional
        Python function to call when the active item changes (user selecting a node) (default is None)
    displayfullname : bool, optional
        If True the nodes will display the full names, id False only the last part splitted by the separator (default is True)
    width : str, optional
        Width of the treeview widget (default is '200px')
    height : str, optional
        Height of the treeview widget (default is '500px')
    elevation : int, optional
        Elevation to assign to the widget (default is 0)
    color : str, optional
        Color to be used as main color of the Treeview widget (default is settings.color_first)
    dark : bool, optional
        If True, the treeview widget will have a dark background (default is settings.dark_mode)
    transition : bool, optional
        If True applies a transition when nodes are opened and closed (default is False)
    selectable : bool, optional
        If True the nodes of the tree have a checkbox to select them (default is True)
    activatable : bool, optional
        If True, one of the nodes of the tree can be activated (default is False)
    active : str, optional
        Name of the node to activate on start (default is None)
    selected : list of str, optional
        List of fullnames of the nodes to select at start (default is []).
    disabled : list of str, optional
        List of fullnames of the nodes to display as disabled in the tree (default is [])
    iconsshow : bool, optional
        If True, an icon is added to each node of the tree (default is False)
    iconscolor : str, optional
        Color of the icons (default is settings.color_first)
    iconsDict: dict, optional
        Dictionary to apply icons to the fullnames of the items, if iconsshow is True (default is None)
    tooltips : bool, optional
        If True the nodes will show the tooltip (default is False)
    tooltips_chars : int, optional
        Minimum lenght of the node label to show the tooltip (default is 20)
    item_height : int, optional
        Item height in pixels (default is 24)
    font_size : int, optional
        Font size in pixels (default is 15)
    icon_size : int, optional
        Icon size in pixels (default is 18)
    checkbox_size : int, optional
        Checkbox size in pixels (default is 24)

    Returns
    -------
    v.Card: ipyvuetify Card widget
        An ipyvuetify v.Card widget having a v.Html as its only child. The v.Html has a treeview.CustomTreeview widget as its only child

    Example
    -------
    Creation and display of a treeview::
    
        from vois.vuetify import treeview
        from IPython.display import display
        
        treecard = treeview.createFlatTreeview(['A','B','C','D','E'], color='green', height='150px')
        display(treecard)

    .. figure:: figures/treeview4.png
       :scale: 100 %
       :alt: flat treeview widget

       Flat treeview widget created from a list of strings.
    """
    
    
    nextid = 1
    roots = []
    names2id = {}
    for name in nameslist:
        elem = {'id': nextid, 'name': name, 'fullname': name}
        if name in disabled: elem['disabled'] = True
        if not iconsDict is None:
            if name in iconsDict:
                elem['icon'] = iconsDict[name]
        roots.append(elem)
        names2id[name] = nextid
        nextid += 1

    selectedids = []
    if select_all: selectedids = [(x+1) for x in range(nextid)]
    else:          selectedids = [names2id[x] for x in selected if x in names2id]
    
    # Convert active name to id
    if not active is None:
        activeid = names2id[active]
    else:
        activeid = None
    
    tree = CustomTreeview(items=roots, selected=selectedids, on_change=on_change, on_activated=on_activated, 
                          expand_selection_to_parents=False, dark=dark, transition=transition, 
                          color=color, selectable=selectable, activatable=activatable, active=activeid, 
                          tooltips=tooltips, tooltips_chars=tooltips_chars,
                          iconsshow=iconsshow, iconscolor=iconscolor,
                          item_height=item_height, font_size=font_size, icon_size=icon_size, checkbox_size=checkbox_size)
    treehtml = v.Html(tag='div', height=height, children=[tree], style_='overflow: hidden;')
    treecard = v.Card(width=width, height=height, elevation=elevation, children=[treehtml],
                      dark=dark, style_='overflow-x: visible;')
    
    return treecard



##################################################################################################################################
# UTILITY FUNCTIONS FOR CREATING THE TREE
##################################################################################################################################

# Add a node to the tree
def addNode(name, parent, icon,
            iconsshow, iconsfolder,
            repeat_parent_as_first_child, rootName, separator,
            nextid, fullnamenodes, fullnames2id,
            disabled):

    # Calculates the fullname of a node
    def getFullname(name, parent):
        if parent: return parent['fullname'] + separator + name
        else:      return name
            
    fullname = getFullname(name,parent)
        
    if fullname in fullnamenodes:
        return fullnamenodes[fullname], nextid, fullnamenodes, fullnames2id
    else:
        elem = {'id': nextid, 'name': name, 'fullname': fullname }
        if (not icon is None) and (len(icon) > 0): elem['icon'] = icon
        if fullname in disabled: elem['disabled'] = True
                
        fullnamenodes[fullname] = elem
        fullnames2id[fullname] = nextid
        nextid += 1
        if parent != None:
            if not 'children' in parent:
                parent['children'] = []
                    
                if repeat_parent_as_first_child and parent['name'] != rootName:
                    firstelem = {'id': nextid, 'name': parent['name'], 'fullname': getFullname(parent['name'],parent) }
                    if parent['fullname'] in disabled: firstelem['disabled'] = True
                    fullnames2id[firstelem['fullname']] = nextid
                    nextid += 1
                    parent['children'].append(firstelem)
                    
            parent['children'].append(elem)
            if iconsshow and iconsfolder: parent['isfolder'] = True
        return elem, nextid, fullnamenodes, fullnames2id

    
# Given a node: returns 
def calculateIdsOfLeaves(node, listofids=[]):
    if not 'children' in node:
        listofids.append(node['id'])
    else:
        for n in node['children']:
            listofids = calculateIdsOfLeaves(n,listofids)
    return listofids
    

    
##################################################################################################################################
# Create a ipyvuetify Treeview from two columns of a DataFrame
# Returns an ipyvuetify widget containing the tree
# Example:
# createTreeviewFromDF2Columns(df, 'col1', 'col2', rootName='A')
##################################################################################################################################
def createTreeviewFromDF2Columns(df,
                                 columnName1,
                                 columnName2,
                                 rootName='Root',
                                 separator='.',
                                 select_all=True,
                                 on_change=None,
                                 on_activated=None,
                                 expand_selection_to_parents=True,
                                 width='200px', height='500px',
                                 elevation=0,
                                 repeat_parent_as_first_child=False,
                                 color=settings.color_first,
                                 dark=settings.dark_mode,
                                 transition=False,
                                 selectable=True,
                                 activatable=False,
                                 active=None,
                                 open_on_click=False,
                                 opened=[],
                                 opened_all=False,
                                 selected=[],
                                 disabled=[],
                                 iconsshow=False,
                                 iconscolor=settings.color_first,
                                 iconsfolder=True,
                                 icons_folder_opened='mdi-folder-open',
                                 icons_folder_closed='mdi-folder',
                                 iconroot=None,
                                 iconscolumnName1=None,
                                 iconscolumnName2=None,
                                 tooltips=False,
                                 tooltips_chars=20,
                                 item_height=24,
                                 font_size=15,
                                 icon_size=18,
                                 checkbox_size=24):
    """
    Create a two levels treeview form the strings contained in two columns of a Pandas DataFrame.

    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame containing at least two columns of type string
    columnName1 : str
        Name of the column of the Pandas DataFrame df containing the parent strings
    columnName2 : str
        Name of the column of the Pandas DataFrame df containing the child strings
    rootName : str, optional
        Name to be displayed as root of the tree (default is 'Root')
    separator : str, optional
        String or character to be considered as separator for calculating the fullname of a node (default is '.'). The full name of a node is the concatenation of the path to reach the node, using the separator character
    select_all : bool, optional
        Flag to control the initial selection of all the nodes of the tree (default is True)
    on_change : function, optional
        Python function to call when the selected nodes change caused by user clicking in one of the checkboxes (default is None). The function receives as argument a list of all the fullnames of the selected nodes
    on_activated : function, optional
        Python function to call when the active item changes (user selecting a node) (default is None)
    expand_selection_to_parents : bool, optional
        If True, also the parent nodes are returned as selected when all children are selected (default is True)
    width : str, optional
        Width of the treeview widget (default is '200px')
    height : int, optional
        Height of the treeview widget (default is '500px')
    elevation : int, optional
        Elevation to assign to the widget (default is 0)
    repeat_parent_as_first_child : bool, optional
        If True each parent node will have a first children with its name (default is False)
    color : str, optional
        Color to be used as main color of the Treeview widget (default is settings.color_first)
    dark : bool, optional
        If True, the treeview widget will have a dark background (default is settings.dark_mode)
    transition : bool, optional
        If True applies a transition when nodes are opened and closed (default is False)
    selectable : bool, optional
        If True the nodes of the tree have a checkbox to select them (default is True)
    activatable : bool, optional
        If True, one of the nodes of the tree can be activated (default is False)
    active : str, optional
        Fullname of the node to activate on start (default is None). The full name of a node is the concatenation of the path to reach the node, using the separator character
    open_on_click : bool, optional
        If True, the nodes of the tree can be opened also by clicking on the node label (default is False)
    opened : list of str, optional
        List of fullnames of the nodes to open at start (default is []). The full name of a node is the concatenation of the path to reach the node, using the separator character
    opened_all : bool, optional
        If True, all the tree nodes are opened at start (default is False). This setting has prevalence over the opened parameter.
    selected : list of str, optional
        List of fullnames of the nodes to select at start (default is []). The full name of a node is the concatenation of the path to reach the node, using the separator character
    disabled : list of str, optional
        List of fullnames of the nodes to display as disabled in the tree (default is [])
    iconsshow : bool, optional
        If True, an icon is added to each node of the tree (default is False)
    iconscolor : str, optional
        Color of the icons (default is settings.color_first)
    iconsfolder : bool, optional
        If True (and if iconsshow is True) it adds the open/closed folder icon to nodes that have children (default is True)
    icons_folder_opened : str, optional
        Name of the icon to use for opened nodes that have children when iconsfolder is True (default 'mdi-folder-open')
    icons_folder_closed : str, optional
        Name of the icon to use for closed nodes that have children when iconsfolder is True (default 'mdi-folder')
    iconroot : str, optional
        Name of the icon for the root node of the tree (default is None)
    iconscolumnName1 : str, optional
        Name of the DataFrame column that contains the icon name for the node on the first level of the tree (default is None)
    iconscolumnName2 : str, optional
        Name of the DataFrame column that contains the icon name for the node on the second level of the tree (default is None)
    tooltips : bool, optional
        If True the nodes will show the tooltip (default is False)
    tooltips_chars : int, optional
        Minimum lenght of the node label to show the tooltip (default is 20)
    item_height : int, optional
        Item height in pixels (default is 24)
    font_size : int, optional
        Font size in pixels (default is 15)
    icon_size : int, optional
        Icon size in pixels (default is 18)
    checkbox_size : int, optional
        Checkbox size in pixels (default is 24)

    Returns
    -------
    v.Card
        An ipyvuetify v.Card widget having a v.Html as its only child. The v.Html has a treeview.CustomTreeview widget as its only child

    Example
    -------
    Creation of a treeview from a simple Pandas DataFrame::
    
        from vois.vuetify import treeview
        import pandas as pd

        table = [['parent', 'child'], ['John', 'Mary'], ['John', 'Peter'], 
                                      ['Ann', 'Hellen'], ['Ann', 'Sue'], ['Ann', 'Claire']]
        headers = table.pop(0)
        df = pd.DataFrame(table, columns=headers)
        display(df)
        
        def on_change(arg):
            print(arg)
            
        treecard = treeview.createTreeviewFromDF2Columns(df, headers[0], headers[1], 
                                                         rootName='Families',
                                                         on_change=on_change)
        display(treecard)
        
    .. figure:: figures/treeview2.png
       :scale: 100 %
       :alt: treeview widget

       Treeview widget created from a Pandas DataFrame.
    """

    nextid = 1          # id of the next node
    fullnamenodes = {}  # All the nodes inserted, key is the fullname of the node
    fullnames2id  = {}  # Convert a fullname to its id
    
    if len(rootName) == 0: rootName = 'Root'
    root, nextid, fullnamenodes, fullnames2id = addNode(rootName,None,iconroot, iconsshow,iconsfolder,repeat_parent_as_first_child,rootName,separator,nextid,fullnamenodes,fullnames2id,disabled)

    # Create the tree from two columns of a df
    for index,row in df.iterrows():
        parentname = str(row[columnName1])
        name       = str(row[columnName2])

        parenticon = None
        if iconsshow and (not iconscolumnName1 is None) and (iconscolumnName1 in row): parenticon = str(row[iconscolumnName1])
        parent, nextid, fullnamenodes, fullnames2id = addNode(parentname, root, parenticon, iconsshow,iconsfolder,repeat_parent_as_first_child,rootName,separator,nextid,fullnamenodes,fullnames2id,disabled)
        
        icon = None
        if iconsshow and (not iconscolumnName2 is None) and (iconscolumnName2 in row): icon = str(row[iconscolumnName2])
        elem, nextid, fullnamenodes, fullnames2id = addNode(name,parent,icon, iconsshow,iconsfolder,repeat_parent_as_first_child,rootName,separator,nextid,fullnamenodes,fullnames2id,disabled)

        
    # List of selected ids
    selectedids = []
    if select_all: selectedids = calculateIdsOfLeaves(root)
    else:          selectedids = [fullnames2id[x] for x in selected if x in fullnames2id]
        
        
    # Convert opened fullnames to ids
    openedids = []
    if not opened_all:
        if len(opened) == 0: openedids = [1]  # Only Root is opened by default
        else:                openedids = [fullnames2id[x] for x in opened if x in fullnames2id]
        
    # Convert active fullname to id
    if not active is None and active in fullnames2id:
        activeid = fullnames2id[active]
    else:
        activeid = None
    
    #import json
    #print(json.dumps(root, indent=4))
    
    tree = CustomTreeview(items=[root], selected=selectedids, opened=openedids, on_change=on_change, on_activated=on_activated, 
                          expand_selection_to_parents=expand_selection_to_parents, dark=dark, transition=transition, opened_all=opened_all, 
                          color=color, selectable=selectable, activatable=activatable, active=activeid, open_on_click=open_on_click,
                          tooltips=tooltips, tooltips_chars=tooltips_chars,
                          iconsshow=iconsshow, iconscolor=iconscolor, icons_folder_opened=icons_folder_opened, icons_folder_closed=icons_folder_closed,
                          item_height=item_height, font_size=font_size, icon_size=icon_size, checkbox_size=checkbox_size)
            
    treehtml = v.Html(tag='div',height=height, children=[tree], style_='overflow: hidden;')
    treecard = v.Card(width=width, height=height, elevation=elevation, children=[treehtml],
                      dark=dark, style_='overflow-x: visible;')
    
    return treecard


##################################################################################################################################
# Create a ipyvuetify Treeview from three columns of a DataFrame
# Returns an ipyvuetify widget containing the tree
# Example:
# createTreeviewFromDF3Columns(df, 'col1', 'col2', 'col3', rootName='A')
##################################################################################################################################
def createTreeviewFromDF3Columns(df,
                                 columnName1,
                                 columnName2,
                                 columnName3,
                                 rootName='Root',
                                 separator='.',
                                 select_all=True,
                                 on_change=None,
                                 on_activated=None,
                                 expand_selection_to_parents=True,
                                 width='200px', height='500px',
                                 elevation=0,
                                 repeat_parent_as_first_child=False,
                                 color=settings.color_first,
                                 dark=settings.dark_mode,
                                 transition=False,
                                 selectable=True,
                                 activatable=False,
                                 active=None,
                                 open_on_click=False,
                                 opened=[],
                                 opened_all=False,
                                 selected=[],
                                 disabled=[],
                                 iconsshow=False,
                                 iconscolor=settings.color_first,
                                 iconsfolder=True,
                                 icons_folder_opened='mdi-folder-open',
                                 icons_folder_closed='mdi-folder',
                                 iconroot=None,
                                 iconscolumnName1=None,
                                 iconscolumnName2=None,
                                 iconscolumnName3=None,
                                 tooltips=False,
                                 tooltips_chars=20,
                                 item_height=24,
                                 font_size=15,
                                 icon_size=18,
                                 checkbox_size=24):
    """
    Create a three levels treeview form the strings contained in three columns of a Pandas DataFrame.

    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame containing at least two columns of type string
    columnName1 : str
        Name of the column of the Pandas DataFrame df containing the parent strings
    columnName2 : str
        Name of the column of the Pandas DataFrame df containing the child strings
    columnName3 : str
        Name of the column of the Pandas DataFrame df containing the sub-child strings
    rootName : str, optional
        Name to be displayed as root of the tree (default is 'Root')
    separator : str, optional
        String or character to be considered as separator for calculating the fullname of a node (default is '.'). The full name of a node is the concatenation of the path to reach the node, using the separator character
    select_all : bool, optional
        Flag to control the initial selection of all the nodes of the tree (default is True)
    on_change : function, optional
        Python function to call when the selected nodes change caused by user clicking in one of the checkboxes (default is None). The function receives as argument a list of all the fullnames of the selected nodes
    on_activated : function, optional
        Python function to call when the active item changes (user selecting a node) (default is None)
    expand_selection_to_parents : bool, optional
        If True, also the parent nodes are returned as selected when all children are selected (default is True)
    width : str, optional
        Width of the treeview widget (default is '200px')
    height : int, optional
        Height of the treeview widget (default is '500px')
    elevation : int, optional
        Elevation to assign to the widget (default is 0)
    repeat_parent_as_first_child : bool, optional
        If True each parent node will have a first children with its name (default is False)
    color : str, optional
        Color to be used as main color of the Treeview widget (default is settings.color_first)
    dark : bool, optional
        If True, the treeview widget will have a dark background (default is settings.dark_mode)
    transition : bool, optional
        If True applies a transition when nodes are opened and closed (default is False)
    selectable : bool, optional
        If True the nodes of the tree have a checkbox to select them (default is True)
    activatable : bool, optional
        If True, one of the nodes of the tree can be activated (default is False)
    active : str, optional
        Fullname of the node to activate on start (default is None). The full name of a node is the concatenation of the path to reach the node, using the separator character
    open_on_click : bool, optional
        If True, the nodes of the tree can be opened also by clicking on the node label (default is False)
    opened : list of str, optional
        List of fullnames of the nodes to open at start (default is []). The full name of a node is the concatenation of the path to reach the node, using the separator character
    opened_all : bool, optional
        If True, all the tree nodes are opened at start (default is False). This setting has prevalence over the opened parameter.
    selected : list of str, optional
        List of fullnames of the nodes to select at start (default is []). The full name of a node is the concatenation of the path to reach the node, using the separator character
    disabled : list of str, optional
        List of fullnames of the nodes to display as disabled in the tree (default is [])
    iconsshow : bool, optional
        If True, an icon is added to each node of the tree (default is False)
    iconscolor : str, optional
        Color of the icons (default is settings.color_first)
    iconsfolder : bool, optional
        If True (and if iconsshow is True) it adds the open/closed folder icon to nodes that have children (default is True)
    icons_folder_opened : str, optional
        Name of the icon to use for opened nodes that have children when iconsfolder is True (default 'mdi-folder-open')
    icons_folder_closed : str, optional
        Name of the icon to use for closed nodes that have children when iconsfolder is True (default 'mdi-folder')
    iconroot : str, optional
        Name of the icon for the root node of the tree (default is None)
    iconscolumnName1 : str, optional
        Name of the DataFrame column that contains the icon name for the node on the first level of the tree (default is None)
    iconscolumnName2 : str, optional
        Name of the DataFrame column that contains the icon name for the node on the second level of the tree (default is None)
    iconscolumnName3 : str, optional
        Name of the DataFrame column that contains the icon name for the node on the third level of the tree (default is None)
    tooltips : bool, optional
        If True the nodes will show the tooltip (default is False)
    tooltips_chars : int, optional
        Minimum lenght of the node label to show the tooltip (default is 20)
    item_height : int, optional
        Item height in pixels (default is 24)
    font_size : int, optional
        Font size in pixels (default is 15)
    icon_size : int, optional
        Icon size in pixels (default is 18)
    checkbox_size : int, optional
        Checkbox size in pixels (default is 24)

    Returns
    -------
    v.Card
        An ipyvuetify v.Card widget having a v.Html as its only child. The v.Html has a treeview.CustomTreeview widget as its only child

    Example
    -------
    Creation of a treeview from a simple Pandas DataFrame::
    
        from vois.vuetify import treeview
        import pandas as pd

        table = [['parent', 'child', 'nephew'], ['John', 'Mary', 'Johnny'], ['John', 'Peter', 'Lisa'], 
                                                ['Ann', 'Hellen', 'July'], ['Ann', 'Sue', 'Hellen'],
                                                ['Ann', 'Claire', 'Pieter']]
        headers = table.pop(0)
        df = pd.DataFrame(table, columns=headers)
        display(df)
        
        def on_change(arg):
            print(arg)
            
        treecard = treeview.createTreeviewFromDF3Columns(df, headers[0], headers[1], headers[2], 
                                                         rootName='Families',
                                                         on_change=on_change)
        display(treecard)
        
    .. figure:: figures/treeview3.png
       :scale: 100 %
       :alt: treeview widget

       Treeview widget created from a Pandas DataFrame.
    """

    
    
    nextid = 1          # id of the next node
    fullnamenodes = {}  # All the nodes inserted, key is the fullname of the node
    fullnames2id  = {}  # Convert a fullname to its id
    
    if len(rootName) == 0: rootName = 'Root'
    root, nextid, fullnamenodes, fullnames2id = addNode(rootName,None,iconroot, iconsshow,iconsfolder,repeat_parent_as_first_child,rootName,separator,nextid,fullnamenodes,fullnames2id,disabled)

    # Create the tree from 3 columns of a df
    for index,row in df.iterrows():
        parentname = str(row[columnName1])
        name       = str(row[columnName2])
        subname    = str(row[columnName3])

        parenticon = None
        if iconsshow and (not iconscolumnName1 is None) and (iconscolumnName1 in row): parenticon = str(row[iconscolumnName1])
        parent, nextid, fullnamenodes, fullnames2id = addNode(parentname, root, parenticon, iconsshow,iconsfolder,repeat_parent_as_first_child,rootName,separator,nextid,fullnamenodes,fullnames2id,disabled)
        
        icon = None
        if iconsshow and (not iconscolumnName2 is None) and (iconscolumnName2 in row): icon = str(row[iconscolumnName2])
        elem, nextid, fullnamenodes, fullnames2id = addNode(name,parent,icon, iconsshow,iconsfolder,repeat_parent_as_first_child,rootName,separator,nextid,fullnamenodes,fullnames2id,disabled)

        icon = None
        if iconsshow and (not iconscolumnName3 is None) and (iconscolumnName3 in row): icon = str(row[iconscolumnName3])
        elem, nextid, fullnamenodes, fullnames2id = addNode(subname,elem,icon, iconsshow,iconsfolder,repeat_parent_as_first_child,rootName,separator,nextid,fullnamenodes,fullnames2id,disabled)
        
        
    # List of selected ids
    selectedids = []
    if select_all: selectedids = calculateIdsOfLeaves(root)
    else:          selectedids = [fullnames2id[x] for x in selected if x in fullnames2id]
        
        
    # Convert opened fullnames to ids
    openedids = []
    if not opened_all:
        if len(opened) == 0: openedids = [1]  # Only Root is opened by default
        else:                openedids = [fullnames2id[x] for x in opened if x in fullnames2id]
        
    # Convert active fullname to id
    if not active is None and active in fullnames2id:
        activeid = fullnames2id[active]
    else:
        activeid = None
    
    #import json
    #print(json.dumps(root, indent=4))
    
    tree = CustomTreeview(items=[root], selected=selectedids, opened=openedids, on_change=on_change, on_activated=on_activated, 
                          expand_selection_to_parents=expand_selection_to_parents, dark=dark, transition=transition, opened_all=opened_all, 
                          color=color, selectable=selectable, activatable=activatable, active=activeid, open_on_click=open_on_click,
                          tooltips=tooltips, tooltips_chars=tooltips_chars,
                          iconsshow=iconsshow, iconscolor=iconscolor, icons_folder_opened=icons_folder_opened, icons_folder_closed=icons_folder_closed,
                          item_height=item_height, font_size=font_size, icon_size=icon_size, checkbox_size=checkbox_size)
    
    treehtml = v.Html(tag='div',height=height, children=[tree], style_='overflow: hidden;')
    treecard = v.Card(width=width, height=height, elevation=elevation, children=[treehtml],
                      dark=dark, style_='overflow-x: visible;')
    
    return treecard



"""Reduced version of vois.colors module: utility functions and classes to manage colors and color interpolation with minimal dependencies."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import math
import random
import numpy as np


# Returns True if the (r,g,b) color id dark
def isColorDark(rgb):
    """
    Returns True if the color (r,g,b) is dark

    Parameters
    ----------
    rgb : tuple
        Tuple of 3 integers representing the RGB components in the range [0,255]
       
    """
    [r,g,b] = rgb
    hsp = math.sqrt(0.299 * (r * r) + 0.587 * (g * g) + 0.114 * (b * b))
    if hsp > 127.5:
        return False
    else:
        return True

# Returns the Multiply blend of two colors strings
def multiply(color1,color2):
    """
    Returns the Multiply blend of two colors strings
    """
    rgb1 = string2rgb(color1)
    rgb2 = string2rgb(color2)
    norm1 = [x/255.0 for x in rgb1]
    norm2 = [x/255.0 for x in rgb2]
    return rgb2hex(tuple([int(255.0*x[0]*x[1]) for x in zip(norm1,norm2)]))


# Returns the Darken blend of two colors strings
def darken(color1,color2):
    """
    Returns the Darken blend of two colors strings
    """
    rgb1 = string2rgb(color1)
    rgb2 = string2rgb(color2)
    return rgb2hex(tuple([min(x[0],x[1]) for x in zip(rgb1,rgb2)]))


# Return a random color in the '#rrggbb' format
def randomColor():
    """
    Returns a random color in the '#rrggbb' format
    """
    r = random.randint(0,255)
    g = random.randint(0,255)
    b = random.randint(0,255)
    return rgb2hex((r,g,b))


# Utility: From (r,g,b) to '#rrggbb'
def rgb2hex(rgb):
    """
    Converts from a color represented as a (r,g,b) tuple to a hexadecimal string representation of the color '#rrggbb'

    Parameters
    ----------
    rgb : tuple of 3 int values
        Input color described by its RGB components as 3 integer values in the range [0,255]
        
    Returns
    -------
        A string containing the color represented as hexadecimals in the '#rrggbb' format
        
    Example
    -------
    Convert a color from (r,g,b) to '#rrggbb'::
    
        from vois import colors
        print( colors.rgb2hex( (255,0,0) ) )
        
    """
    return '#{:02x}{:02x}{:02x}'.format(rgb[0], rgb[1], rgb[2])


# Utility: From '#rrggbb' to (r,g,b)
def hex2rgb(color):
    """
    Converts from a hexadecimal string representation of the color '#rrggbb' to a (r,g,b) tuple

    Parameters
    ----------
    color : string
        A string containing the color represented as hexadecimals in the '#rrggbb' format
        
    Returns
    -------
        Tuple of 3 integers representing the RGB components in the range [0,255]
        
    Example
    -------
    Convert a color from '#rrggbb' to (r,g,b)::

        from vois import colors
        print( colors.hex2rgb( '#ff0000' ) )
        
    """
    if color[0] == '#':
        color = color[1:]
    rgb = (int(color[0:2],16), int(color[2:4],16), int(color[4:6],16))
    return rgb


# Utility: From 'rgb(a,b,c)' to (r,g,b)
def text2rgb(color):
    """
    Converts from string representation of the color 'rgb(r,g,b)' to a (r,g,b) tuple

    Parameters
    ----------
    color : string
        A string containing the color represented in the 'rgb(r,g,b)' format
        
    Returns
    -------
        Tuple of 3 integers representing the RGB components in the range [0,255]
        
    Example
    -------
    Convert a color from 'rgb(r,g,b)' to (r,g,b)::
    
        from vois import colors
        print( colors.text2rgb( 'rgb(255,0,0)' ) )
        
    """
    if color[0:4] == 'rgb(':
        rgb = color[4:].replace(')','').split(',')
        if len(rgb) >= 3:
            return ((int(rgb[0]),int(rgb[1]),int(rgb[2])))
           
    return (0,0,0)


# Utility: Convert a color string in '#rrggbb' or in 'rgb(...)' format into a tuple (r,g,b)
def string2rgb(s):
    """
    Converts from string representation of the color 'rgb(r,g,b)' or '#rrggbb' to a (r,g,b) tuple

    Parameters
    ----------
    color : string
        A string containing the color represented in the 'rgb(r,g,b)' format or in the '#rrggbb' format
        
    Returns
    -------
        Tuple of 3 integers representing the RGB components in the range [0,255]
    """
    if s[0] == '#':
        return hex2rgb(s)
    elif s[0:4] == 'rgb(':
        return text2rgb(s)
    return (0,0,0)



# colorInterpolator class
class colorInterpolator:
    """
    Class to perform color interpolation given a list of colors and a numerical range [minvalue,maxvalue]. The list of colors is considered as a linear range spanning from minvalue to maxvalue and the method :py:meth:`colors.colorInterpolator.GetColor` can be used to calculate any intermediate color by passing as input any numeric value.

    Parameters
    ----------
    colorlist : list of strings representing colors in 'rgb(r,g,b)' or '#rrggbb' format
        Input list of colors
    minvalue: float, optional
        Minimum value for the interpolation (default is 0.0)
    maxvalue: float, optional
        Maximum numerical value for the interpolation (default is 100.0)
        
    Examples
    --------
    Creation of a color interpolator from a list of custom colors::
    
        from vois import colors
        
        colorlist = ['rgb(247,251,255)',
                     'rgb(198,219,239)',
                     'rgb(107,174,214)', 
                     'rgb(33,113,181)',
                     'rgb(8,48,107)']
        c = colors.colorInterpolator(colorlist)
        print( c.GetColor(50.0) )
    
    Creation of a color interpolator using one of the Plotly library predefined colorscales (see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )::
    
        import plotly.express as px
        from vois import colors
        
        c = colors.colorInterpolator(px.colors.sequential.Viridis, 0.0, 100.0)
        print( c.GetColor(33.3) )
               
               
    To visualize a color palette from a list of colors, the :py:func:`colors.paletteImage` function can be used::
    
        from vois import colors
        import plotly.express as px

        img = colors.paletteImage(px.colors.sequential.Blues, width=400, height=40)
        display(img)

    .. figure:: figures/paletteImage.png
       :scale: 100 %
       :alt: Plotly colorscale

       Display of a Plotly colorscale.
    """

    # Initialization
    def __init__(self, colorlist, minValue=0.0, maxValue=100.0):

        self.minValue = minValue
        self.maxValue = maxValue
        
        self.colors = [string2rgb(color) for color in colorlist]
        
                    
    # Return '#rrggbb' color linearly interpolated 
    def GetColor(self, value):
        """
        Returns a color in the '#rrggbb' format linearly interpolated in the [minvalue,maxvalue] range
        
        Parameters
        ----------
        value : float
            Numeric value for which the color has to be calculated
                
        Returns
        -------
        A string containing the color represented as hexadecimals in the '#rrggbb' format
            
        """
        if self.minValue >= self.maxValue:  # Avoid division by zero!
            return rgb2hex(self.palette[-1])
        
        if value < self.minValue: value = self.minValue
        if value > self.maxValue: value = self.maxValue
        
        n = len(self.colors)
        
        a = list(np.linspace(self.minValue, self.maxValue, n))
        i2 = next(x[0] for x in enumerate(a) if x[1] >= value)
        i1 = i2 - 1
        
        d1 = abs(value - a[i1])
        d2 = abs(a[i2] - value)
        d = d1 + d2
        w1 = 1.0 - d1/d
        w2 = 1.0 - w1
        
        c1 = self.colors[i1]
        c2 = self.colors[i2]

        r = int(c1[0]*w1 + c2[0]*w2)
        g = int(c1[1]*w1 + c2[1]*w2)
        b = int(c1[2]*w1 + c2[2]*w2)
       
        return rgb2hex((r,g,b))
        
        
    # Interpolate colors and returns a list of num_classes colors
    def GetColors(self, num_classes):
        """
        Returns a list of colors in the '#rrggbb' format covering all the input colors
        
        Parameters
        ----------
        num_classes : int
            Number of colors to interpolate
                
        Returns
        -------
        A list of strings containing the colors represented as hexadecimals in the '#rrggbb' format
        """
        if num_classes >= 2:
            return [self.GetColor(perc) for perc in np.linspace(self.minValue, self.maxValue, num_classes)]
        
        return self.colors
               
        
        
    # repr
    def __repr__(self):
        s = [str(x) for x in self.colors]
        return '-'.join(s)



"""Vertically aligned list of customizable cards with items that can be moved, added and removed."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from ipywidgets import widgets, Layout
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import tooltip
except:
    import settings
    import tooltip


#####################################################################################################################################################
# Vertically aligned list of customizable cards with items that can be moved, added and removed.
#####################################################################################################################################################
class sortableList:
    """
    Vertically aligned list of customizable cards with items that can be moved, added and removed.
                
    Parameters
    ----------
    items : list of dicts, optional
        List of items to display in the list (default is [])
    width : int, optional
        Width of widget in pixels (default is 400)
    outlined : bool, optional
        Flag to show each of the items with a border (default is True)
    dark : bool, optional
        Flag to invert the text and backcolor (default is the value of settings.dark_mode)
    allowNew : bool, optional
        If True, a 'plus' button is displayed that allows for adding new items (default is True)
    newOnTop: bool, optional
        If True, the '+' button adds a new item in first position on the items list (default False, new items are added as last in the items list)
    newButtonOnTop: bool, optional
        If True, the '+' button is displayed on top of the first item, otherwise it is displayed below the last item (default is False, the '+' button is on the bottom)
    itemNew : function, optional
        Python function called when a new items is added. The function is called with no arguments and it must return the dict initialized with the new item content (default is None). As an alternative, the function can return None, but then the real adding of the new item must be done by directly calling the doAddItem method
    itemContent : function, optional
        Python function called when an item is displayed. The function is called with an item as its first argument and the index (position of the item) as second argument. The function must return a list containing the ipyvuetify widgets to display the item content (default is None)
    bottomContent : list of ipyvuetify widgets, optional
        Additional widgets content to display in the bottom line (containing the 'plus' button), aligned to the right (default is [])
    onchange : function, optional
        Python function to call when the user changes the order of items or removes an item. The function will receive no parameters as input (default is None)
    onmovedown : function, optional
        Python function to call when the user moves one item down. The function will receive as parameter the zero-based index, before the move, of the moved item (default is None)
    onmoveup : function, optional
        Python function to call when the user moves one item up. The function will receive as parameter the zero-based index, before the move, of the moved item (default is None)
    onremoving : function, optional
        Python function to call when the user is about to remove an item. The function will receive as parameter the zero-based index of the item that is going to be removed (default is None)
    onremoved : function, optional
        Python function to call just after the user removes an item. The function will receive as parameter the zero-based index of the removed item (default is None)
    onadded : function, optional
        Python function to call just after a new item is added. The function will receive as parameter the zero-based index of the new item (default is None)
    buttonstooltip : bool, optional
        If True, the buttons to mode, add, remove items will have a tooltip (default is False)
    tooltipadd : str, optional
        Tooltip text for the "add" button (default is 'Add new')
    tooltipdown : str, optional
        Tooltip text for the "move down" buttons (default is 'Move down')
    tooltipup : str, optional
        Tooltip text for the "move up" buttons (default is 'Move up')
    tooltipremove : str, optional
        Tooltip text for the "remove" buttons (default is 'Remove')
    activatable : bool, optional
        If True the items can be activated by clicking on them (default is False)
    ondeactivated : function, optional
        Python function to call just after an item that was the active one, is deactivated. The function will receive as parameter the zero-based index of the deactivated item (default is None)
    onactivated : function, optional
        Python function to call just after an item becomes the active one (or by user-clicking or by setting the active property). The function will receive as parameter the zero-based index of the active item (default is None)


    Examples
    --------
    Simple list displaying static text::
        
        from vois.vuetify import sortableList
        from ipywidgets import widgets
        from IPython.display import display
        import ipyvuetify as v

        items = [{ "name": 'Jane Adams',   "email": 'jane@adams.com'   },
                 { "name": 'Paul Davis',   "email": 'paul@davis.com'   },
                 { "name": 'Amanda Brown', "email": 'amanda@brown.com' }
                ]


        # Creation of a new item
        def itemNew():
            return {"name": "new", "email": "empty"}


        # Content of an item
        def itemContent(item, index):
            return [
                v.CardSubtitle(class_="mb-n4", children=[item['name']]),
                v.CardText(    class_="mt-n2", children=[item['email']])
            ]

        s = sortableList.sortableList(items=items, dark=False, allowNew=True,
                                      itemNew=itemNew, itemContent=itemContent)
        display(s.draw())
    
    
    .. figure:: figures/sortableList1.png
       :scale: 100 %
       :alt: label widget

       Example of a simple sortableList displaying static text


    Example of a sortable list displaying editable text, boolean and date values (using the :py:class:`datePicker.datePicker` class)::
        
        from vois.vuetify import sortableList, datePicker, switch, tooltip
        from ipywidgets import widgets
        from IPython.display import display
        import ipyvuetify as v

        output = widgets.Output()

        items = [
            { "name": "Paul",    "surname": "Dockery",  "married": False, "date": "" },
            { "name": "July",    "surname": "Winters",  "married": True,  "date": "1997-07-28" },
            { "name": "David",   "surname": "Forest",   "married": True,  "date": "1999-03-03" },
            { "name": "Dorothy", "surname": "Landmann", "married": False, "date": "" }
        ]

        dark = False

        # Called when an item is moved or deleted
        def onchange():
            with output:
                print('Changed!')


        # Creation of a new item
        def itemNew():
            return { "name": "", "surname": "", "married": False, "date": "" }


        # Remove all items
        def itemRemoveAll(widget, event, data):
            s.items = []

        reset = v.Btn(icon=True, children=[v.Icon(children=['mdi-playlist-remove'])])
        reset.on_event('click', itemRemoveAll)


        # Content of an item
        def itemContent(item, index):

            def onname(widget, event, data):
                item["name"] = int(data)

            def onsurname(widget, event, data):
                item["surname"] = data

            def onmarried(flag):
                item["married"] = flag
                dp.disabled = not flag
                if not flag:
                    item["date"] = ''
                    dp.date = None

            def ondate():
                item["date"] = dp.date

            tfname = v.TextField(label='Name:', value=item['name'],
                                 color='amber', dense=True,
                                 style_="max-width: 70px", class_="pa-0 ma-0 mt-2")
            tfname.on_event('input', onname)

            tfsurname = v.TextField(label='Surname:', value=item['surname'],
                                    color='amber', dense=True, 
                                    style_="max-width: 100px", class_="pa-0 ma-0 mt-2")
            tfsurname.on_event('input', onsurname)

            sw = switch.switch(item['married'], "Married", onchange=onmarried)

            dp = datePicker.datePicker(date=item['date'], dark=dark, width=88, 
                                       onchange=ondate, offset_x=True, offset_y=False)
            dp.disabled = not item['married']

            sp = v.Html(tag='div', class_="pa-0 ma-0 mr-3", children=[''])

            return [ v.Row(class_="pa-0 ma-0 ml-2", no_gutters=True,
                           children=[tfname, sp, tfsurname, sp, sw.draw(), sp, dp.draw()]) ]


        s = sortableList.sortableList(items=items,
                                      width=520,
                                      outlined=False,
                                      dark=dark,
                                      allowNew=True,
                                      itemNew=itemNew,
                                      itemContent=itemContent,
                                      bottomContent=[tooltip.tooltip("Remove all persons",
                                                                     reset)],
                                      onchange=onchange,
                                      buttonstooltip=True)
        display(s.draw())
        display(output)

    .. figure:: figures/sortableList2.png
       :scale: 100 %
       :alt: label widget

       Example of a sortableList to edit textual, boolean and date values on persons.
    """

   
    # Initialization
    def __init__(self, items=[], width=400, maxheightlist=10000, outlined=True, dark=settings.dark_mode,
                 allowNew=True, newOnTop=False, newButtonOnTop=False, itemNew=None, itemContent=None, bottomContent=[],
                 onchange=None, onmovedown=None, onmoveup=None, onremoving=None, onremoved=None, onadded=None, buttonstooltip=False,
                 tooltipadd='Add new', tooltipdown='Move down', tooltipup='Move up', tooltipremove='Remove',
                 activatable=False, ondeactivated=None, onactivated=None):
        
        self._items         = items
        self.width          = width
        self.outlined       = outlined
        self.dark           = dark

        self.allowNew       = allowNew
        self.newOnTop       = newOnTop
        self.newButtonOnTop = newButtonOnTop
        self.itemNew        = itemNew
        self.itemContent    = itemContent
        
        self.bottomContent  = bottomContent
        
        self.onchange       = onchange
        self.buttonstooltip = buttonstooltip
        self.tooltipadd     = tooltipadd
        self.tooltipdown    = tooltipdown
        self.tooltipup      = tooltipup
        self.tooltipremove  = tooltipremove
        self.onmovedown     = onmovedown
        self.onmoveup       = onmoveup
        self.onremoving     = onremoving
        self.onremoved      = onremoved
        self.onadded        = onadded
        
        self.activatable    = activatable
        self.ondeactivated  = ondeactivated
        self.onactivated    = onactivated
        self.activeindex    = -1
        
        self.output     = v.Card(flat=True, max_height='%dpx'%maxheightlist, children=[])
        self.outputplus = v.Card(flat=True, children=[])

        self.cards    = []
        self.bdowns   = []
        self.bups     = []
        self.bremoves = []

        # Create the + button to add a new item
        self.style = "min-width: %dpx; max-width: %dpx" %(self.width,self.width)
        if self.allowNew:
            self.plusbutton = v.Btn(icon=True, dark=self.dark, children=[v.Icon(children=['mdi-plus'])])
            self.plusbutton.on_event('click', self.onadd)
            if self.buttonstooltip:
                self.outputplus.children = [v.Row(no_gutters=True, style_=self.style, justify='space-between',
                                                  children=[tooltip.tooltip(self.tooltipadd, self.plusbutton)] + self.bottomContent)]
            else:
                self.outputplus.children = [v.Row(no_gutters=True, style_=self.style, justify='space-between',
                                                  children=[self.plusbutton] + self.bottomContent)]
            

        # Add all the items
        for index, item in enumerate(self._items):
            self.additem(item, index==0, index==len(self._items)-1)

        if self.activatable: bottomspace = 'mb-2'
        else:                bottomspace = ''
        self.col = v.Col(class_="pa-0 ma-0 %s"%bottomspace, children=self.cards)
        self.output.children = [self.col]
        

    # Click on + button
    def onadd(self, widget, event, data):
        if self.itemNew:
            item = self.itemNew()
            if not item is None:
                self.doAddItem(item)
                if not self.onadded is None:
                    if self.newOnTop:
                        self.onadded(0)
                    else:
                        self.onadded(len(self.cards)-1)

                        
    # Effective add of a new item
    def doAddItem(self, item):
        """
        Manual adding of a new item
        """
        if self.newOnTop:
            if len(self.bups) > 0: self.bups[0].disabled = False
            self._items.insert(0,item)

            if self.activeindex >= 0:
                self.activeindex += 1
            
            self.additem(item, True, len(self.cards)==0, onTop=True)
        else:
            if len(self.bdowns) > 0: self.bdowns[-1].disabled = False
            self._items.append(item)
            self.additem(item, len(self.cards)==0, True)
                
        if self.activatable: bottomspace = 'mb-2'
        else:                bottomspace = ''
        self.col = v.Col(class_="pa-0 ma-0 %s"%bottomspace, children=self.cards)
        self.output.children = [self.col]
        if self.onchange:
            self.onchange()


    # Update the disabled state of the buttons
    def update_buttons(self):
        for i in range(len(self.bdowns)): self.bdowns[i].disabled  = False
        if len(self.bdowns)>0:            self.bdowns[-1].disabled = True
        for i in range(len(self.bups)):   self.bups[i].disabled = False
        if len(self.bups)>0:              self.bups[0].disabled = True


    # Move item down
    def ondown(self, widget, event, data):
        index = self.bdowns.index(widget)
        
        wasactive = False
        if index == self.activeindex:
            wasactive = True
            
        changewithactive = False
        if index+1 == self.activeindex:
            changewithactive = True
            
        a = index
        b = index + 1
        self.updateCard(self._items[a], a)
        self.updateCard(self._items[b], b)
        self._items[b],   self._items[a]   = self._items[a],   self._items[b]
        self.cards[b],    self.cards[a]    = self.cards[a],    self.cards[b]
        self.bdowns[b],   self.bdowns[a]   = self.bdowns[a],   self.bdowns[b]
        self.bups[b],     self.bups[a]     = self.bups[a],     self.bups[b]
        self.bremoves[b], self.bremoves[a] = self.bremoves[a], self.bremoves[b]
        self.update_buttons()
        if self.activatable: bottomspace = 'mb-2'
        else:                bottomspace = ''
        self.col = v.Col(class_="pa-0 ma-0 %s"%bottomspace, children=self.cards)
        self.output.children = [self.col]
        
        if self.onchange:
            self.onchange()
        if self.onmovedown:
            self.onmovedown(index)
            
        if wasactive:
            self.active = index + 1
        elif changewithactive:
            self.active = index


    # Move item up
    def onup(self, widget, event, data):
        index = self.bups.index(widget)
        
        wasactive = False
        if index == self.activeindex:
            wasactive = True
            
        changewithactive = False
        if index-1 == self.activeindex:
            changewithactive = True
            
        a = index - 1
        b = index
        self.updateCard(self._items[a], a)
        self.updateCard(self._items[b], b)
        self._items[b],   self._items[a]   = self._items[a],   self._items[b]
        self.cards[b],    self.cards[a]    = self.cards[a],    self.cards[b]
        self.bdowns[b],   self.bdowns[a]   = self.bdowns[a],   self.bdowns[b]
        self.bups[b],     self.bups[a]     = self.bups[a],     self.bups[b]
        self.bremoves[b], self.bremoves[a] = self.bremoves[a], self.bremoves[b]
        self.update_buttons()
        if self.activatable: bottomspace = 'mb-2'
        else:                bottomspace = ''
        self.col = v.Col(class_="pa-0 ma-0 %s"%bottomspace, children=self.cards)
        self.output.children = [self.col]
        
        if self.onchange:
            self.onchange()
        if self.onmoveup:
            self.onmoveup(index)

        if wasactive:
            self.active = index - 1
        elif changewithactive:
            self.active = index

            
    # Remove item
    def ondel(self, widget, event, data):
        index = self.bremoves.index(widget)
        if self.onremoving:
            self.onremoving(index)
            
        if index == self.activeindex:
            if not self.ondeactivated is None:
                self.ondeactivated(self.activeindex)
            self.activeindex = -1
        elif index <= self.activeindex:
            self.activeindex -= 1
            
        del self._items[index]
        del self.cards[index]
        del self.bdowns[index]
        del self.bups[index]
        del self.bremoves[index]
        self.col.children = self.col.children[:index] + self.col.children[index+1 :]
        self.update_buttons()
        self.output.children = [self.col]
        
        if self.onchange:
            self.onchange()
        if self.onremoved:
            self.onremoved(index)


    # Update the card for an item given its index
    def updateCard(self, item, index):
        if self.itemContent:
            c = self.col.children[index]
            c.children = [c.children[0]] + self.itemContent(item,index)
        
    
    # Management of click on an item card
    def __internal_onclick(self, widget=None, event=None, data=None):
        if widget in self.cards:
            index = self.cards.index(widget)
            if self.activatable:
                self.active = index
        
    
    # Add an item to the lists 
    def additem(self, item, isfirst, islast, onTop=False):
        if self.itemContent:
            bdown   = v.Btn(icon=True, class_="mr-n3", disabled=islast,  children=[v.Icon(small=True, children=['mdi-arrow-down'])])
            bup     = v.Btn(icon=True, class_="mr-n3", disabled=isfirst, children=[v.Icon(small=True, children=['mdi-arrow-up'])])
            bremove = v.Btn(icon=True, children=[v.Icon(small=True, children=['mdi-close'])])

            if onTop:
                self.bdowns.insert(0,bdown)
                self.bups.insert(0,bup)
                self.bremoves.insert(0,bremove)
            else:
                self.bdowns.append(bdown)
                self.bups.append(bup)
                self.bremoves.append(bremove)

            bdown.on_event('click.stop', self.ondown)
            bup.on_event('click.stop', self.onup)
            bremove.on_event('click.stop', self.ondel)

            if self.buttonstooltip:
                buttons = [tooltip.tooltip(self.tooltipdown,   bdown),
                           tooltip.tooltip(self.tooltipup,     bup),
                           tooltip.tooltip(self.tooltipremove, bremove)]
            else:
                buttons = [bdown,bup,bremove]
                
            if self.activatable: ripple = True
            else:                ripple = False
                
            if onTop:
                index = 0
            else:
                index = len(self.cards)
                
            if self.activatable:
                bottom = "mb-2"
            else:
                bottom = ""
                
            c = v.Card(outlined=self.outlined, dark=self.dark, flat=True, dense=True, class_="pa-0 ma-0 mt-1 %s"%bottom, style_=self.style, ripple=ripple, raised=False,
                       children=[v.CardTitle(class_="justify-end pa-0 ma-0 mt-n1 mb-n5", children=buttons)] + self.itemContent(item,index))
            c.on_event('click', self.__internal_onclick)

            if onTop:
                self.cards.insert(0,c)
            else:
                self.cards.append(c)

    
    
    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (a v.Html object displaying two output widgets)"""
        if self.allowNew:
            if self.newButtonOnTop:
                return v.Html(tag='div',children=[widgets.VBox([self.outputplus,self.output])])
            else:
                return v.Html(tag='div',children=[widgets.VBox([self.output,self.outputplus])])
        else:
            return self.output
    
    
    # Get the items
    @property
    def items(self):
        """
        Get/Set the updated items.
        
        Returns
        --------
        items : list of dicts
            List of items in their updated position

        """
        return self._items
            
    # Set the items
    @items.setter
    def items(self, items):
        self._items = items
        
        self.cards    = []
        self.bdowns   = []
        self.bups     = []
        self.bremoves = []
        
        self.activeindex = -1

        # Add all the items
        for index, item in enumerate(self._items):
            self.additem(item, index==0, index==len(self._items)-1)

        if self.activatable: bottomspace = 'mb-2'
        else:                bottomspace = ''
        self.col = v.Col(class_="pa-0 ma-0 %s"%bottomspace, children=self.cards)
        self.output.children = [self.col]
            
        if self.onchange:
            self.onchange()

    # Get the index of the active item
    @property
    def active(self):
        """
        Get/Set the active item.
        
        Returns
        --------
        index : int
            index of the active item

        """
        return self.activeindex
            
    # Set the active item
    @active.setter
    def active(self, index):
        if index >= 0 and index < len(self.cards):
            if self.activeindex >= 0:
                self.cards[self.activeindex].raised = False
                if not self.ondeactivated is None:
                    self.ondeactivated(self.activeindex)

            self.activeindex = index
            self.cards[self.activeindex].raised = True
            if not self.onactivated is None:
                self.onactivated(self.activeindex)
            


"""Widget to select independent options using a list of buttons displayed horizontally or vertically."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import tooltip
    from . import Button
except:
    import settings
    import tooltip
    import Button


#####################################################################################################################################################
# Multi switch  control
#####################################################################################################################################################
class multiSwitch():
    """
    Widget to select independent options using a list of buttons displayed horizontally or vertically.
        
    Parameters
    ----------
    values : list of bool
        Initial state of the buttons representing the independent options
    labels : list of strings
        Strings to be displayed as text of the options
    tooltips : list of strings, optional
        Tooltip text for the options
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    dark : bool, optional
        Flag to invert the text and backcolor (default is the value of settings.dark_mode)
    onchange : function, optional
        Python function to call when the user clicks on one of the buttons. The function will receive a parameter of list containing the status of all the buttons, from 0 to len(labels)-1
    row : bool, optional
        Flag to display the buttons horizontally or vertically (default is True)
    width : int, optional
        Width in pixels of the buttons (default is 150)
    height : int, optional
        Height in pixels of the buttons (default is 36)
    justify : str, optional
        In case of horizontal placement, applies the justify-content css property. Available options are: start, center, end, space-between and space-around.
    rounded : bool, optional
        Flag to display the buttons with a rounded shape (default is the button_rounded flag defined in the settings.py module)
    outlined : bool, optional
        Flag to display the buttons as outlined (default is True)
    colorselected : str, optional
        Color used for the buttons when they are selected (default is settings.color_first)
    colorunselected : str, optional
        Color used for the buttons when they are not selected (default is settings.color_second)
    managedblclick : bool, optional
        If True the dblclick event is managed to select a single button of the multi-switch (default is False)
    paddingrow : int, optional
        Horizontal padding among toggle buttons (1 unit means 4 pixels). Default is 1
    paddingcol : int, optional
        Vertical padding among toggle buttons (1 unit means 4 pixels). Default is 2
    tile : bool, optional
        Flag to remove the buttons small border (default is False)
    large : bool, optional
        Flag that sets the large version of the button (default is False)
    xlarge : bool, optional
        Flag that sets the xlarge version of the button (default is False)
    small : bool, optional
        Flag that sets the small version of the button (default is False)
    xsmall : bool, optional
        Flag that sets the xsmall version of the button (default is False)

    Example
    -------
    Creation and display of a widget for the selection of 3 independent options::
        
        from vois.vuetify import multiSwitch
        from ipywidgets import widgets
        from IPython.display import display
        
        output = widgets.Output()

        def onchange(values):
            with output:
                print(values)

        m = multiSwitch.multiSwitch([False, True, False], ['Option 1', 'Option 2', 'Option 3'],
                                    tooltips=['Tooltip for option 1'], 
                                    onchange=onchange, row=False, width=150, rounded=False, outlined=True,
                                    colorselected='#FFA300', colorunselected='#aaaaaa')

        display(m.draw())
        display(output)


    .. figure:: figures/multiSwitch.png
       :scale: 100 %
       :alt: multiSwitch widget

       multiSwitch widget for selecting independent options using buttons.
   """

    # Initialization
    def __init__(self, values, labels, tooltips=None, color=settings.color_first, onchange=None, dark=settings.dark_mode,
                 row=True, width=150, height=36, justify='space-between', rounded=settings.button_rounded, outlined=False,
                 colorselected=settings.color_first, colorunselected=settings.color_second,
                 managedblclick=False, paddingrow=1, paddingcol=2, tile=False,
                 small=False, xsmall=False, large=False, xlarge=False):
        
        self.values   = [bool(x) for x in values]    # list of boolean values
        self.labels   = labels
        self.tooltips = tooltips
        self.color    = color
        self.dark     = dark
        self.onchange = onchange
        self.row      = row
        self.width    = width
        self.height   = height
        self.justify  = justify
        self.rounded  = rounded
        self.outlined = outlined
        self.colorselected   = colorselected
        self.colorunselected = colorunselected
        self.managedblclick  = managedblclick
        self.paddingrow = paddingrow
        self.paddingcol = paddingcol
        self.tile   = tile
        self.small  = small
        self.xsmall = xsmall
        self.large  = large
        self.xlarge = xlarge
        
        self.__createButtons()
        
        
    # Create the toggle buttons
    def __createButtons(self):

        self.buttons = []
        i = 0
        for label in self.labels:
            tooltip = ''
            if i < len(self.tooltips):
                tooltip = self.tooltips[i]
            
            if self.row: c = "pa-0 ma-0 mr-%d"%self.paddingrow
            else:
                if i == len(self.labels)-1:
                    c = "pa-0 ma-0 mb-7"
                else:
                    c = "pa-0 ma-0 mb-%d"%self.paddingcol
                    
            b = Button(label, class_=c, onclick=self.__internal_onchange, ondblclick=self.__internal_dblclick, argument=i, tooltip=tooltip, selected=self.values[i],
                              small=self.small, xsmall=self.xsmall, large=self.large, xlarge=self.xlarge, width=self.width, height=self.height,
                              rounded=self.rounded, tile=self.tile, outlined=self.outlined, dark=self.dark, colorselected=self.colorselected, colorunselected=self.colorunselected)
            self.buttons.append(b)
            i += 1

        if self.row: self.group = v.Row(class_="pa-0 ma-0", justify=self.justify, children=[x.draw() for x in self.buttons], style_="overflow: hidden;")
        else:        self.group = v.Col(cols=12, class_="pa-0 ma-0", children=[x.draw() for x in self.buttons], style_="overflow: hidden;")


    # Get the active button
    @property
    def value(self):
        """
        Get/Set the status of the multiSwitch buttons.
        
        Returns
        --------
        flags : list of booleans
            Selected status of all the options

        Example
        -------
        Programmatically set the options and print the status of the multiSwitch buttons::
            
            t.value = [False, True, True]
            print(t.value)
        
        """
        return self.values
   
    
    # Set the status of the buttons
    @value.setter
    def value(self, values):
        if len(values) == len(self.values):
            self.values = [bool(x) for x in values]
            for i in range(len(self.values)):
                self.buttons[i].selected = self.values[i]
            if self.onchange:
                self.onchange(self.values)


    # Manage onchange event
    def __internal_onchange(self, index):
        self.values[index] = not self.values[index]
        self.buttons[index].selected = self.values[index]
        if self.onchange:
            self.onchange(self.values)
            
        
    # Manage dblclick event
    def __internal_dblclick(self, index):
        if self.managedblclick:

            i = 0
            for b in self.buttons:
                if i == index: 
                    self.values[i] = True
                    b.selected = True
                else:
                    self.values[i] = False
                    b.selected = False
                    
                i += 1
                
            if self.onchange:
                self.onchange(self.values)
            
    
    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Row or v.Col widget)"""
        return self.group

